"""This module implements the ArrayStats class which calculates several
descriptive statistics of 2, 3 or 4D input arrays. That is, if the input array
is a 4D image (rows, columns, slices, "time"), it calculates statistics for
each 2D slice, each 3D volume and the entire 4D volume. See docstring of the
`calculate` method for the list of the calculated statistical measures.

"""

import numpy as np
from scipy import stats

NOT_CALCULATED_MSG = 'Not calculated. See ArrayStats.calculate().'

class ArrayStats():
    """Class to calculate array statistics (optionally within a mask)

    Parameters
    ----------
    image : np.ndarray
        array with 2, 3 or 4 dimensions
    roi : np.ndarray (dtype: bool), optional
        region of interest, same dimensions as `image`

    Attributes
    ----------
    image : see above (parameters)
    roi : see above (parameters)
    image_shape : tuple
        shape (length of each dimension of image)
    image_ndims: int
        number of dimensions of image

    """
    def __init__(self, image, roi=None):
        """ Init method: see class docstring for parameters/attributes

        """
        image_shape = image.shape
        image_ndims = len(image_shape)

        # Error checks
        if image_ndims < 2 or image_ndims > 4:
            raise ValueError("`image` must be [2, 3, 4]D")

        # Initialise unspecified input arguments
        if roi is None:
            roi = np.ones(image_shape, dtype=bool)
        elif isinstance(roi, np.ndarray) and roi.dtype == bool:
            if roi.shape != image_shape:
                raise ValueError("`roi.shape` must match `image.shape`")
        else:
            raise TypeError("`roi` must None or a numpy array with dtype=bool")

        self.image = image
        self.roi = roi
        self.image_shape = image_shape
        self.image_ndims = image_ndims

    def calculate(self):
        """Calculate array statistics

        Returns
        -------
        dict
            dictionary where the keys are the calculated statistical measures:
            - 'n'        : number of array elements
            - 'mean'
            - 'median'
            - 'min'
            - 'max'
            - 'std'      : standard deviation
            - 'cv'       : coefficient of variation (std/mean)
            - 'skewness'
            - 'kurtosis'
            - 'entropy'
            Each of the statistical measures is a dictionary in itself, where
            keys are the dimensions over which the calculation was performed.
            For example:
            - statistics['mean']['2D']: mean of each 2D slice
            - statistics['mean']['3D']: mean of each 3D volume
            - statistics['mean']['4D']: mean of each 4D volume

        """
        # Init attribute that may need to be modified (i.e. add new axis)
        image = self.image
        roi = self.roi

        # Add new axis to `image` and `roi` if needed
        if self.image_ndims == 2:
            image = image[:, :, np.newaxis, np.newaxis]
            roi = roi[:, :, np.newaxis, np.newaxis]
        elif self.image_ndims == 3:
            image = image[:, :, :, np.newaxis]
            roi = roi[:, :, :, np.newaxis]

        # Get number of slices and time points of expanded `image` (i.e. after
        # np.newaxis). Note the `image_shape` attribute initialised in __init__
        # refers to the shape before adding new axis
        (_, _, nz, nt) = image.shape

        # Pre-allocate 2D
        tmp2 = np.full((nz, nt), np.nan)
        n2 = np.copy(tmp2)
        mean2 = np.copy(tmp2)
        median2 = np.copy(tmp2)
        min2 = np.copy(tmp2)
        max2 = np.copy(tmp2)
        std2 = np.copy(tmp2)
        cv2 = np.copy(tmp2)
        skewness2 = np.copy(tmp2)
        kurtosis2 = np.copy(tmp2)
        entropy2 = np.copy(tmp2)

        # Pre-allocate 3D
        tmp3 = np.full(nt, np.nan)
        n3 = np.copy(tmp3)
        mean3 = np.copy(tmp3)
        median3 = np.copy(tmp3)
        min3 = np.copy(tmp3)
        max3 = np.copy(tmp3)
        std3 = np.copy(tmp3)
        cv3 = np.copy(tmp3)
        skewness3 = np.copy(tmp3)
        kurtosis3 = np.copy(tmp3)
        entropy3 = np.copy(tmp3)

        # Pre-allocate 4D
        tmp4 = np.full(1, np.nan)
        n4 = np.copy(tmp4)
        mean4 = np.copy(tmp4)
        median4 = np.copy(tmp4)
        min4 = np.copy(tmp4)
        max4 = np.copy(tmp4)
        std4 = np.copy(tmp4)
        cv4 = np.copy(tmp4)
        skewness4 = np.copy(tmp4)
        kurtosis4 = np.copy(tmp4)
        entropy4 = np.copy(tmp4)

        # Calculate statistics of the 4D volume
        intensities = image[roi]
        array_stats4 = FlatStats(intensities).calculate()
        n4 = array_stats4.n
        mean4 = array_stats4.mean
        median4 = array_stats4.median
        min4 = array_stats4.min
        max4 = array_stats4.max
        std4 = array_stats4.std
        cv4 = array_stats4.cv
        skewness4 = array_stats4.skewness
        kurtosis4 = array_stats4.kurtosis
        entropy4 = array_stats4.entropy

        for it in range(nt):
            # Calculate statistics of each 3D volume
            it_intensities = image[:, :, :, it][roi[:, :, :, it]]
            array_stats3 = FlatStats(it_intensities).calculate()
            n3[it] = array_stats3.n
            mean3[it] = array_stats3.mean
            median3[it] = array_stats3.median
            min3[it] = array_stats3.min
            max3[it] = array_stats3.max
            std3[it] = array_stats3.std
            cv3[it] = array_stats3.cv
            skewness3[it] = array_stats3.skewness
            kurtosis3[it] = array_stats3.kurtosis
            entropy3[it] = array_stats3.entropy

            for iz in range(nz):
                # Calculate statistics of each 2D slice
                iz_intensities = image[:, :, iz, it][roi[:, :, iz, it]]
                array_stats2 = FlatStats(iz_intensities).calculate()
                n2[iz, it] = array_stats2.n
                mean2[iz, it] = array_stats2.mean
                median2[iz, it] = array_stats2.median
                min2[iz, it] = array_stats2.min
                max2[iz, it] = array_stats2.max
                std2[iz, it] = array_stats2.std
                cv2[iz, it] = array_stats2.cv
                skewness2[iz, it] = array_stats2.skewness
                kurtosis2[iz, it] = array_stats2.kurtosis
                entropy2[iz, it] = array_stats2.entropy

        if self.image_ndims == 4:
            # Init dict for each dimension of each statistic
            n = {
                '2D': n2,  # number of voxels in each 2D slice
                '3D': n3,  # number of voxels in each 3D volume
                '4D': n4   # number of voxels in 4D volume
            }
            mean = {
                '2D': mean2,  # mean of each 2D slice
                '3D': mean3,  # mean of each 3D volume
                '4D': mean4   # mean of the 4D volume
            }
            median = {
                '2D': median2,
                '3D': median3,
                '4D': median4
            }
            minimum = {
                '2D': min2,
                '3D': min3,
                '4D': min4
            }
            maximum = {
                '2D': max2,
                '3D': max3,
                '4D': max4
            }
            std = {
                '2D': std2,
                '3D': std3,
                '4D': std4
            }
            cv = {
                '2D': cv2,
                '3D': cv3,
                '4D': cv4
            }
            skewness = {
                '2D': skewness2,
                '3D': skewness3,
                '4D': skewness4
            }
            kurtosis = {
                '2D': kurtosis2,
                '3D': kurtosis3,
                '4D': kurtosis4
            }
            entropy = {
                '2D': entropy2,
                '3D': entropy3,
                '4D': entropy4
            }
        elif self.image_ndims == 3:
            n = {
                '2D': n2.transpose()[0],
                '3D': n4, # n4 because {statistic}4 always returns the result
            }             # over the entire array, which here is 3D
            mean = {
                '2D': mean2.transpose()[0],
                '3D': mean4,
            }
            median = {
                '2D': median2.transpose()[0],
                '3D': median4,
            }
            minimum = {
                '2D': min2.transpose()[0],
                '3D': min4,
            }
            maximum = {
                '2D': max2.transpose()[0],
                '3D': max4,
            }
            std = {
                '2D': std2.transpose()[0],
                '3D': std4,
            }
            cv = {
                '2D': cv2.transpose()[0],
                '3D': cv4,
            }
            skewness = {
                '2D': skewness2.transpose()[0],
                '3D': skewness4,
            }
            kurtosis = {
                '2D': kurtosis2.transpose()[0],
                '3D': kurtosis4,
            }
            entropy = {
                '2D': entropy2.transpose()[0],
                '3D': entropy4,
            }
        else:
            n = n4             # n4 because {statistic}4 always returns the
            mean = mean4       # result over the entire array, which here is 2D
            median = median4
            minimum = min4
            maximum = max4
            std = std4
            cv = cv4
            skewness = skewness4
            kurtosis = kurtosis4
            entropy = entropy4

        # Init statistics "wrapper" dictionary
        statistics = {
            'n': n,
            'mean': mean,
            'median': median,
            'min': minimum,
            'max': maximum,
            'std': std,
            'cv': cv,
            'skewness': skewness,
            'kurtosis': kurtosis,
            'entropy': entropy
        }

        return statistics


class FlatStats():
    """Helper class that calculates statistics of a flat (1D) array

    Parameters
    ----------
    x : np.ndarray
        flat array (1 dimension)

    Attributes
    ----------
    x : see above (parameters)
    n : float
        number of array elements
    mean : float
    median : float
    min : float
    max : float
    std : float
        standard deviation
    cv : float
        coefficient of variation (std/mean)
    skewness : float
    kurtosis : float
    entropy : float

    """
    def __init__(self, x):
        """ Init method: see class documentation for parameters/attributes

        """
        if np.ndim(x) != 1:
            raise ValueError("`x` should be a flat (1D) array")

        self.x = x

        # Init statistics (calculated in calculate())
        self.n = NOT_CALCULATED_MSG
        self.mean = NOT_CALCULATED_MSG
        self.median = NOT_CALCULATED_MSG
        self.min = NOT_CALCULATED_MSG
        self.max = NOT_CALCULATED_MSG
        self.std = NOT_CALCULATED_MSG
        self.cv = NOT_CALCULATED_MSG
        self.skewness = NOT_CALCULATED_MSG
        self.kurtosis = NOT_CALCULATED_MSG
        self.entropy = NOT_CALCULATED_MSG

    def calculate(self):
        """Calculate flat array statistics

        Returns
        -------
        FlatStats object with calculated statistics

        """

        if self.x is None or self.x.size == 0:
            n = 0
            mean = np.nan
            median = np.nan
            minimum = np.nan
            maximum = np.nan
            std = np.nan
            cv = np.nan
            skewness = np.nan
            kurtosis = np.nan
            entropy = np.nan
        elif np.isnan(self.x).any():
            n = len(self.x)
            mean = np.nan
            median = np.nan
            minimum = np.nan
            maximum = np.nan
            std = np.nan
            cv = np.nan
            skewness = np.nan
            kurtosis = np.nan
            entropy = np.nan
        else:
            n = len(self.x)
            mean = np.mean(self.x)
            median = np.median(self.x)
            minimum = np.min(self.x)
            maximum = np.max(self.x)
            std = np.std(self.x)
            if mean == 0:
                cv = np.nan
            else:
                cv = std/mean

            # Don't need to deal with `nan_policy` as we catch arrays with nans
            # above and assign nans to the resulting statistical metrics
            skewness = stats.skew(self.x, bias=True)
            kurtosis = stats.kurtosis(self.x, fisher=True, bias=True)
            entropy = stats.entropy(self.x)

        self.n = n
        self.mean = mean
        self.median = median
        self.min = minimum
        self.max = maximum
        self.std = std
        self.cv = cv
        self.skewness = skewness
        self.kurtosis = kurtosis
        self.entropy = entropy

        return self



import concurrent.futures
import nibabel as nib
import numpy as np
import os
import warnings
from tqdm import tqdm
from scipy.optimize import curve_fit


class T1:
    """
    Attributes
    ----------
    t1_map : np.ndarray
        The estimated T1 values in ms
    t1_err : np.ndarray
        The certainty in the fit of `t1` in ms
    m0_map : np.ndarray
        The estimated M0 values
    m0_err : np.ndarray
        The certainty in the fit of `m0`
    eff_map : np.ndarray
        The estimated inversion efficiency where 0 represents no inversion
        pulse and 2 represents a 180 degree inversion
    eff_err : np.ndarray
        The certainty in the fit of `eff`
    shape : tuple
        The shape of the T1 map
    n_ti : int
        The number of TI used to calculate the map
    """

    def __init__(self, pixel_array, inversion_list, affine, tss=0, tss_axis=-2,
                 mask=None, parameters=2, molli=False, multithread=True):
        """Initialise a T1 class instance.

        Parameters
        ----------
        pixel_array : np.ndarray
            A array containing the signal from each voxel at each inversion
            time with the last dimension being time i.e. the array needed to
            generate a 3D T1 map would have dimensions [x, y, z, TI].
        inversion_list : list()
            An array of the inversion times used for the last dimension of the
            raw data. In milliseconds.
        tss : float, optional
            Default 0
            The temporal slice spacing is the delay between acquisition of
            slices in a T1 map. Including this information means the
            inversion time is correct for each slice in a multi-slice T1
            map. In milliseconds.
        tss_axis : int, optional
            Default -2 i.e. last spatial axis
            The axis over which the temporal slice spacing is applied. This
            axis is relative to the full 4D pixel array i.e. tss_axis=-1
            would be along the TI axis and would be meaningless.
            If `pixel_array` is single slice (dimensions [x, y, TI]),
            then this should be set to None.
        affine : np.ndarray
            A matrix giving the relationship between voxel coordinates and
            world coordinates.
        mask : np.ndarray, optional
            A boolean mask of the voxels to fit. Should be the shape of the
            desired T1 map rather than the raw data i.e. omit the time
            dimension.
        parameters : {2, 3}, optional
            Default `2`
            The number of parameters to fit the data to. A two parameter fit
            will estimate S0 and T1 while a three parameter fit will also
            estimate the inversion efficiency.
        molli : bool, optional
            Default False.
            Apply MOLLI corrections to T1.
        multithread : bool, optional
            Default True.
            If True, fitting will be distributed over all cores available on
            the node. If False, fitting will be carried out on a single thread.
            Multithreading is useful when calculating the T1 for a large
            number of voxels e.g. generating a multi-slice abdominal T1 map.
            Turning off multithreading can be useful when fitting very small
            amounts of data e.g. a mean T1 signal decay over a ROI when the
            overheads of multithreading are more of a hindrance than the
            increase in speed distributing the calculation would generate.
        """

        self.pixel_array = pixel_array
        self.shape = pixel_array.shape[:-1]
        self.dimensions = len(pixel_array.shape)
        self.n_ti = pixel_array.shape[-1]
        self.affine = affine
        # Generate a mask if there isn't one specified
        if mask is None:
            self.mask = np.ones(self.shape, dtype=bool)
        else:
            self.mask = mask
        # Don't process any nan values
        self.mask[np.isnan(np.sum(pixel_array, axis=-1))] = False
        self.inversion_list = inversion_list
        self.tss = tss
        if tss_axis is not None:
            self.tss_axis = tss_axis % self.dimensions
        else:
            self.tss_axis = None
            self.tss = 0
        self.parameters = parameters
        self.molli = molli
        self.multithread = multithread

        # Some sanity checks
        assert (pixel_array.shape[-1]
                == len(inversion_list)), 'Number of inversions does not ' \
                                         'match the number of time frames ' \
                                         'on the last axis of pixel_array'
        if self.tss != 0:
            assert (self.tss_axis != self.dimensions - 1), \
                'Temporal slice spacing can\'t be applied to the TI axis.'
            assert (tss_axis < self.dimensions), \
                'tss_axis must be less than the number of spatial dimensions'
        if self.molli:
            if self.parameters == 2:
                self.parameters = 3
                warnings.warn('MOLLI requires a three parameter fit, '
                              'using parameters=3.')


        # Initialise output attributes
        self.t1_map = np.zeros(self.shape)
        self.t1_err = np.zeros(self.shape)
        self.m0_map = np.zeros(self.shape)
        self.m0_err = np.zeros(self.shape)
        self.eff_map = np.zeros(self.shape)
        self.eff_err = np.zeros(self.shape)

        # Fit data
        if self.parameters == 2:
            self.t1_map, self.t1_err, self.m0_map, self.m0_err = self.__fit__()
        elif self.parameters == 3:
            self.t1_map, self.t1_err, self.m0_map, self.m0_err, \
                self.eff_map, self.eff_err = self.__fit__()
        else:
            raise ValueError('Parameters can be 2 or 3 only. You specified '
                             '{}'.format(self.parameters))

        if self.molli:
            correction_factor = (self.m0_map * self.eff_map) / self.m0_map - 1
            percentage_error = self.t1_err / self.t1_map
            self.t1_map = np.nan_to_num(self.t1_map * correction_factor)
            self.t1_err = np.nan_to_num(self.t1_map * percentage_error)

    def __fit__(self):
        n_vox = np.prod(self.shape)
        # Initialise maps
        t1_map = np.zeros(n_vox)
        m0_map = np.zeros(n_vox)
        t1_err = np.zeros(n_vox)
        m0_err = np.zeros(n_vox)
        if self.parameters == 3:
            eff_map = np.zeros(n_vox)
            eff_err = np.zeros(n_vox)
        mask = self.mask.flatten()
        signal = self.pixel_array.reshape(-1, self.n_ti)
        slices = np.indices(self.shape)[self.tss_axis].ravel()
        # Get indices of voxels to process
        idx = np.argwhere(mask).squeeze()

        # Multithreaded method
        if self.multithread:
            with concurrent.futures.ProcessPoolExecutor() as pool:
                with tqdm(total=idx.size) as progress:
                    futures = []

                    for ind in idx:
                        ti_slice_corrected = self.inversion_list + \
                                             slices[ind] * self.tss
                        future = pool.submit(self.__fit_signal__,
                                             signal[ind, :],
                                             ti_slice_corrected,
                                             self.parameters)
                        future.add_done_callback(lambda p: progress.update())
                        futures.append(future)

                    results = []
                    for future in futures:
                        result = future.result()
                        results.append(result)

            if self.parameters == 2:
                t1_map[idx], t1_err[idx], \
                    m0_map[idx], m0_err[idx] = [np.array(row)
                                                for row in zip(*results)]
            elif self.parameters == 3:
                t1_map[idx], t1_err[idx], \
                    m0_map[idx], m0_err[idx], \
                        eff_map[idx], eff_err[idx] = [np.array(row)
                                                      for row in zip(*results)]

        # Single threaded method
        else:
            with tqdm(total=idx.size) as progress:
                for ind in idx:
                    sig = signal[ind, :]
                    ti_slice_corrected = self.inversion_list + \
                                            slices[ind] * self.tss
                    if self.parameters == 2:
                        t1_map[ind], t1_err[ind], \
                            m0_map[ind], m0_err[ind] = \
                                self.__fit_signal__(sig,
                                                    ti_slice_corrected,
                                                    self.parameters)
                    elif self.parameters == 3:
                        t1_map[ind], t1_err[ind], \
                            m0_map[ind], m0_err[ind], \
                                eff_map[ind], eff_err[ind] = \
                                    self.__fit_signal__(sig,
                                                        ti_slice_corrected,
                                                        self.parameters)
                    progress.update(1)

        # Reshape results to raw data shape
        t1_map = t1_map.reshape(self.shape)
        m0_map = m0_map.reshape(self.shape)
        t1_err = t1_err.reshape(self.shape)
        m0_err = m0_err.reshape(self.shape)

        if self.parameters == 2:
            return t1_map, t1_err, m0_map, m0_err

        elif self.parameters == 3:
            eff_map = eff_map.reshape(self.shape)
            eff_err = eff_err.reshape(self.shape)
            return t1_map, t1_err, m0_map, m0_err, eff_map, eff_err

    def __fit_signal__(self, sig, t, parameters):

        # Initialise parameters and specify equation to fit to
        if parameters == 2:
            bounds = ([0, 0], [5000, 10000000])
            initial_guess = [1000, 30000]
            if sig.min() >= 0:
                eq = two_param_abs_eq
            else:
                eq = two_param_eq
        elif parameters == 3:
            bounds = ([0, 0, 1], [5000, 10000000, 2])
            initial_guess = [1000, 30000, 2]
            if sig.min() >= 0:
                eq = three_param_abs_eq
            else:
                eq = three_param_eq

        # Fit data to equation
        try:
            popt, pcov = curve_fit(eq, t, sig,
                                   p0=initial_guess, bounds=bounds)
        except RuntimeError:
            popt = np.zeros(self.parameters)
            pcov = np.zeros((self.parameters, self.parameters))

        # Extract fits and errors from result variable
        if popt[0] < bounds[1][0] - 1:
            t1 = popt[0]
            m0 = popt[1]
            err = np.sqrt(np.diag(pcov))
            t1_err = err[0]
            m0_err = err[1]
            if self.parameters == 3:
                eff = popt[2]
                eff_err = err[2]
        else:
            t1, m0, t1_err, m0_err = 0, 0, 0, 0
            if self.parameters == 3:
                eff, eff_err = 0, 0

        if self.parameters == 2:
            return t1, t1_err, m0, m0_err
        elif self.parameters == 3:
            return t1, t1_err, m0, m0_err, eff, eff_err

    def r1_map(self):
        """
        Generates the R1 map from the T1 map output by initialising this
        class.

        Parameters
        ----------
        See class attributes in __init__

        Returns
        -------
        r1_map : np.ndarray
            An array containing the R1 map generated
            by the function with R1 measured in ms.
        """
        return np.nan_to_num(np.reciprocal(self.t1_map), posinf=0, neginf=0)

    def to_nifti(self, output_directory=os.getcwd(), base_file_name='Output',
                 maps='all'):
        """Exports some of the T1 class attributes to NIFTI.

        Parameters
        ----------
        output_directory : string, optional
            Path to the folder where the NIFTI files will be saved.
        base_file_name : string, optional
            Filename of the resulting NIFTI. This code appends the extension.
            Eg., base_file_name = 'Output' will result in 'Output.nii.gz'.
        maps : list or 'all', optional
            List of maps to save to NIFTI. This should either the string "all"
            or a list of maps from ["t1", "t1_err", "m0", "m0_err", "eff",
            "eff_err", "r1", "mask"]
        """
        os.makedirs(output_directory, exist_ok=True)
        base_path = os.path.join(output_directory, base_file_name)
        if maps == 'all' or maps == ['all']:
            maps = ['t1', 't1_err', 'm0', 'm0_err', 'eff', 'eff_err', 'r1_map',
                    'mask']
        if isinstance(maps, list):
            for result in maps:
                if result == 't1' or result == 't1_map':
                    t1_nifti = nib.Nifti1Image(self.t1_map, affine=self.affine)
                    nib.save(t1_nifti, base_path + '_t1_map.nii.gz')
                elif result == 't1_err':
                    t1_err_nifti = nib.Nifti1Image(self.t1_err,
                                                   affine=self.affine)
                    nib.save(t1_err_nifti, base_path + '_t1_err.nii.gz')
                elif result == 'm0' or result == 'm0_map':
                    m0_nifti = nib.Nifti1Image(self.m0_map, affine=self.affine)
                    nib.save(m0_nifti, base_path + '_m0_map.nii.gz')
                elif result == 'm0_err':
                    m0_err_nifti = nib.Nifti1Image(self.m0_err,
                                                   affine=self.affine)
                    nib.save(m0_err_nifti, base_path + '_m0_err.nii.gz')
                elif (self.parameters == 3) and \
                     (result == 'eff' or result == 'eff_map'):
                    eff_nifti = nib.Nifti1Image(self.eff_map,
                                                affine=self.affine)
                    nib.save(eff_nifti, base_path + '_eff_map.nii.gz')
                elif self.parameters == 3 and result == 'eff_err':
                    eff_err_nifti = nib.Nifti1Image(self.eff_err,
                                                    affine=self.affine)
                    nib.save(eff_err_nifti, base_path + '_eff_err.nii.gz')
                elif result == 'r1' or result == 'r1_map':
                    r1_nifti = nib.Nifti1Image(T1.r1_map(self),
                                               affine=self.affine)
                    nib.save(r1_nifti, base_path + '_r1_map.nii.gz')
                elif result == 'mask':
                    mask_nifti = nib.Nifti1Image(self.mask.astype(int),
                                                 affine=self.affine)
                    nib.save(mask_nifti, base_path + '_mask.nii.gz')
        else:
            raise ValueError('No NIFTI file saved. The variable "maps" '
                             'should be "all" or a list of maps from '
                             '"["t1", "t1_err", "m0", "m0_err", "eff", '
                             '"eff_err", "r1", "mask"]".')

        return


def two_param_abs_eq(t, t1, m0):
    """
    Calculate the expected signal from the equation signal = abs(M0 * (1 -
    2 * exp(-t / T1)))

    Parameters
    ----------
    t: list
        The times the signal will be calculated at
    t1: float
        The T1 of the signal
    m0: float
        The M0 of the signal

    Returns
    -------
    signal: ndarray
    """
    return np.abs(m0 * (1 - 2 * np.exp(-t / t1)))


def two_param_eq(t, t1, m0):
    """
    Calculate the expected signal from the equation signal = M0 * (1 - 2 *
    exp(-t / T1))

    Parameters
    ----------
    t: list
        The times the signal will be calculated at
    t1: float
        The T1 of the signal
    m0: float
        The M0 of the signal

    Returns
    -------
    signal: ndarray
    """
    return m0 * (1 - 2 * np.exp(-t / t1))


def three_param_abs_eq(t, t1, m0, eff):
    """
    Calculate the expected signal from the equation signal = abs(M0 * (1 -
    eff * exp(-t / T1)))

    Parameters
    ----------
    t: list
        The times the signal will be calculated at
    t1: float
        The T1 of the signal
    m0: float
        The M0 of the signal
    eff: float
        The inversion efficiency (where 0 is no inversion and 2 is a 180
        degree inversion)

    Returns
    -------
    signal: ndarray
    """
    return np.abs(m0 * (1 - eff * np.exp(-t / t1)))


def three_param_eq(t, t1, m0, eff):
    """
    Calculate the expected signal from the equation signal = M0 * (1 - eff *
    exp(-t / T1)))

    Parameters
    ----------
    t: list
        The times the signal will be calculated at
    t1: float
        The T1 of the signal
    m0: float
        The M0 of the signal
    eff: float
        The inversion efficiency (where 0 is no inversion and 2 is a 180
        degree inversion)

    Returns
    -------
    signal: ndarray
    """
    return m0 * (1 - eff * np.exp(-t / t1))


def magnitude_correct(pixel_array):
    """Sign corrects the magnitude of inversion recovery data using the
    complex component of the signal.

    This function uses the methods of Jerzy Szumowski et al
    (https://doi.org/10.1002/jmri.23705).

    Parameters
    ----------
    pixel_array: ndarray
        Can either be a complex array or have the real and imaginary
        parts of the image as the final dimension e.g. a complex 3D image
        could have the dimensions [x, y, z, ti] where [0, 0, 0, 0] = 1 + 2j
        or the dimensions [x, y, z, ti, type] where [0, 0, 0, 0, 0] = 1 and
        [0, 0, 0, 0, 1] = 2.

    Returns
    -------
    corrected_array : ndarray
        An array of the magnitude intensities with signs corrected.
    """

    # Convert data to a complex array if it isn't already
    if not np.iscomplexobj(pixel_array):
        if pixel_array.shape[-1] == 2:
            pixel_array = pixel_array[..., 0] + pixel_array[..., 1] * (0 + 1j)
        else:
            raise ValueError('Last axis of pixel_array must have length 2')

    pixel_array_prime = np.zeros(pixel_array.shape, dtype=np.complex128)

    for ti in range(pixel_array.shape[-1]):
        pixel_array_prime[..., ti] = (pixel_array[..., ti] *
                                      pixel_array[..., -1].conjugate()) \
                                     / np.abs(pixel_array[..., -1])

    phase_factor = np.imag(np.log(pixel_array_prime / np.abs(pixel_array)))
    phase_offset = np.abs(phase_factor) - (np.pi / 2)
    sign = -(phase_offset / np.abs(phase_offset))
    corrected_array = sign * np.abs(pixel_array)
    return corrected_array



import os
import numpy as np
import nibabel as nib
import warnings


class MTR:
    """
    Generates a Magnetisation Transfer Ratio (MTR) map from a series of volumes
    collected with 2 different magnetisation transfer values (ON and OFF).

    Attributes
    ----------
    mtr_map : np.ndarray
        The estimated magnetisation transfer ratio values.
        Each value should be a float between 0 and 1.
    shape : tuple
        The shape of the MTR map.
    mt_on : np.ndarray
        The array corresponding to the magnetisation transfer value ON.
    mt_off : np.ndarray
        The array corresponding to the magnetisation transfer value OFF.
    mask : np.ndarray
        A boolean mask of the voxels to fit.
    """

    def __init__(self, pixel_array, affine, mask=None):
        """Initialise a MTR class instance.

        Parameters
        ----------
        pixel_array : np.ndarray
            A 3D/4D array containing the image at each magnetisation transfer
            value, i.e. the dimensions of the array are (x, y, 2) or
            (x, y, z, 2). The assumption is that `mt_off` comes first in
            `pixel_array` and `mt_on` comes second in `pixel_array`, i.e.
            `pixel_array[..., 0]` corresponds to `mt_off` and
            `pixel_array[..., 1]` corresponds to `mt_on`.
        affine : np.ndarray
            A matrix giving the relationship between voxel coordinates and
            world coordinates.
        mask : np.ndarray, optional
            A boolean mask of the voxels to fit. Should be the shape of the
            desired MTR map rather than the raw data i.e. omit the on/off
            dimension.
        """
        assert (pixel_array.shape[-1] == 2), 'The input should contain 2 MT ' \
                                             'values (ON / OFF). The last ' \
                                             'dimension of the input ' \
                                             'pixel_array must be 2.'
        if np.sum(pixel_array[..., 1]) >= np.sum(pixel_array[..., 0]):
            warnings.warn(f'The average intensity of the MT_ON image is more '
                          'than the average intensity of the MT_OFF image. '
                          'This will lead to negative MTR values which is not '
                          'usually desirable. Please check that you\'ve input '
                          'MT_ON and MT_OFF in the correct order in '
                          '`pixel_array`.')
        self.pixel_array = pixel_array
        self.shape = pixel_array.shape[:-1]
        self.affine = affine
        # Generate a mask if there isn't one specified
        if mask is None:
            self.mask = np.ones(self.shape, dtype=bool)
        else:
            self.mask = mask
        # The assumption is that MT_OFF comes first in `pixel_array`
        self.mt_off = self.pixel_array[..., 0] * self.mask
        # The assumption is that MT_ON comes second in `pixel_array`
        self.mt_on = self.pixel_array[..., 1] * self.mask
        # Magnetisation Transfer Ratio calculation
        self.mtr_map = np.nan_to_num(((self.mt_off - self.mt_on) /
                                     self.mt_off), posinf=0, neginf=0)

    def to_nifti(self, output_directory=os.getcwd(), base_file_name='Output',
                 maps='all'):
        """Exports some of the MTR class attributes to NIFTI.

        Parameters
        ----------
        output_directory : string, optional
            Path to the folder where the NIFTI files will be saved.
        base_file_name : string, optional
            Filename of the resulting NIFTI. This code appends the extension.
            Eg., base_file_name = 'Output' will result in 'Output.nii.gz'.
        maps : list or 'all', optional
            List of maps to save to NIFTI. This should either the string "all"
            or a list of maps from ["mtr_map", "mt_on", "mt_off", "mask"].
        """
        os.makedirs(output_directory, exist_ok=True)
        base_path = os.path.join(output_directory, base_file_name)
        if maps == 'all' or maps == ['all']:
            maps = ['mtr_map', 'mt_on', 'mt_off', 'mask']
        if isinstance(maps, list):
            for result in maps:
                if result == 'mtr' or result == 'mtr_map':
                    mtr_nifti = nib.Nifti1Image(self.mtr_map,
                                                affine=self.affine)
                    nib.save(mtr_nifti, base_path + '_mtr_map.nii.gz')
                elif result == 'mt_on':
                    mt_on_nifti = nib.Nifti1Image(self.mt_on,
                                                  affine=self.affine)
                    nib.save(mt_on_nifti, base_path + '_mt_on.nii.gz')
                elif result == 'mt_off':
                    mt_off_nifti = nib.Nifti1Image(self.mt_off,
                                                   affine=self.affine)
                    nib.save(mt_off_nifti, base_path + '_mt_off.nii.gz')
                elif result == 'mask':
                    mask_nifti = nib.Nifti1Image(self.mask.astype(int),
                                                 affine=self.affine)
                    nib.save(mask_nifti, base_path + '_mask.nii.gz')
        else:
            raise ValueError('No NIFTI file saved. The variable "maps" '
                             'should be "all" or a list of maps from '
                             '"["mtr_map", "mt_on", "mt_off", "mask"]".')

        return



import os
import numpy as np
import nibabel as nib
from skimage.restoration import unwrap_phase
from ukat.utils.tools import convert_to_pi_range


class B0:
    """
    Generates a B0 map from a series of volumes collected
    with 2 different echo times.

    Attributes
    ----------
    b0_map : np.ndarray
        The estimated B0 values in Hz
    shape : tuple
        The shape of the B0 map
    mask : np.ndarray
        A boolean mask of the voxels to fit
    n_te : int
        The number of TE used to calculate the map
    delta_te : float
        The difference between the second and the first Echo Time
    phase0 : np.ndarray
        The phase image corresponding to the first Echo Time
    phase1 : np.ndarray
        The phase image corresponding to the second Echo Time
    phase_difference : np.ndarray
        The difference between the 2 phase images
    """

    def __init__(self, pixel_array, echo_list, affine, mask=None,
                 unwrap=True, wrap_around=False):
        """Initialise a B0 class instance.

        Parameters
        ----------
        pixel_array : np.ndarray
            A 4D/3D array containing the phase image at each
            echo time i.e. the dimensions of the array are
            [x, y, TE] or [x, y, z, TE].
        echo_list : list
            An array of the echo times in ms used for the last dimension of the
            raw data.
        affine : np.ndarray
            A matrix giving the relationship between voxel coordinates and
            world coordinates.
        mask : np.ndarray, optional
            A boolean mask of the voxels to fit. Should be the shape of the
            desired B0 map rather than the raw data i.e. omit the echo times
            dimension.
        unwrap : boolean, optional
            By default, this script applies the
            scipy phase unwrapping for each phase echo image.
        wrap_around : boolean, optional
            By default, this flag from unwrap_phase is False.
            The algorithm will regard the edges along the corresponding axis
            of the image to be connected and use this connectivity to guide the
            phase unwrapping process.Eg., voxels [0, :, :] are considered to be
            next to voxels [-1, :, :] if wrap_around=True.
        """

        self.pixel_array = pixel_array
        self.shape = pixel_array.shape[:-1]
        self.n_te = pixel_array.shape[-1]
        self.affine = affine
        # Generate a mask if there isn't one specified
        if mask is None:
            self.mask = np.ones(self.shape, dtype=bool)
        else:
            self.mask = mask
        if self.n_te == len(echo_list) and self.n_te == 2:
            # Get the Echo Times
            echo0 = echo_list[0]
            echo1 = echo_list[1]
            # Calculate DeltaTE in seconds
            self.delta_te = np.absolute(echo1 - echo0) * 0.001
            # Extract each phase image, mask them and rescale to
            # [-pi, pi] if not in that range already.
            self.phase0 = np.ma.masked_array(
                            convert_to_pi_range(np.squeeze(
                                self.pixel_array[..., 0])), mask=mask)
            self.phase1 = np.ma.masked_array(
                            convert_to_pi_range(np.squeeze(
                                self.pixel_array[..., 1])), mask=mask)
            if unwrap:
                # Unwrap each phase image
                self.phase0 = unwrap_phase(self.phase0,
                                           wrap_around=wrap_around)
                self.phase1 = unwrap_phase(self.phase1,
                                           wrap_around=wrap_around)
            # Phase difference
            self.phase_difference = self.phase1 - self.phase0
            # B0 Map calculation
            self.b0_map = self.phase_difference / (2 * np.pi * self.delta_te)
            # The following step checks in a central bounding box containing
            # the kidneys if there is an offset in the B0 Map and corrects
            # that offset if it exists. This is due to a jump in phase
            # that can occurr during the unwrapping of the phase images.
            # Bounding Box
            prop = 0.5
            bx0 = int(self.shape[0] / 2 - prop / 2 * self.shape[0])
            bxf = int(self.shape[0] / 2 + prop / 2 * self.shape[0])
            by0 = int(self.shape[1] / 2 - prop / 2 * self.shape[1])
            byf = int(self.shape[1] / 2 + prop / 2 * self.shape[1])
            # B0 Map Mean inside the bounding box accross the slices
            mean_central_b0 = np.mean(self.b0_map[bx0:bxf, by0:byf, ...])
            # B0 Map Offset Step
            b0_offset_step = 1 / self.delta_te
            # B0 Map Offset Correction
            self.b0_map -= (mean_central_b0 // b0_offset_step) * b0_offset_step
        else:
            raise ValueError('The input should contain 2 echo times.'
                             'The last dimension of the input pixel_array must'
                             'be 2 and the echo_list must only have 2 values.')

    def to_nifti(self, output_directory=os.getcwd(), base_file_name='Output',
                 maps='all'):
        """Exports some of the B0 class attributes to NIFTI.

        Parameters
        ----------
        output_directory : string, optional
            Path to the folder where the NIFTI files will be saved.
        base_file_name : string, optional
            Filename of the resulting NIFTI. This code appends the extension.
            Eg., base_file_name = 'Output' will result in 'Output.nii.gz'.
        maps : list or 'all', optional
            List of maps to save to NIFTI. This should either the string "all"
            or a list of maps from ["b0", "mask", "phase0", "phase1",
            "phase_difference"].
        """
        os.makedirs(output_directory, exist_ok=True)
        base_path = os.path.join(output_directory, base_file_name)
        if maps == 'all' or maps == ['all']:
            maps = ['b0', 'mask', 'phase0', 'phase1', 'phase_difference']
        if isinstance(maps, list):
            for result in maps:
                if result == 'b0' or result == 'b0_map':
                    b0_nifti = nib.Nifti1Image(self.b0_map, affine=self.affine)
                    nib.save(b0_nifti, base_path + '_b0_map.nii.gz')
                elif result == 'mask':
                    mask_nifti = nib.Nifti1Image(self.mask.astype(int),
                                                 affine=self.affine)
                    nib.save(mask_nifti, base_path + '_mask.nii.gz')
                elif result == 'phase0':
                    phase0_nifti = nib.Nifti1Image(self.phase0,
                                                   affine=self.affine)
                    nib.save(phase0_nifti, base_path + '_phase0.nii.gz')
                elif result == 'phase1':
                    phase1_nifti = nib.Nifti1Image(self.phase1,
                                                   affine=self.affine)
                    nib.save(phase1_nifti, base_path + '_phase1.nii.gz')
                elif result == 'phase_difference':
                    phase_diff_nifti = nib.Nifti1Image(self.phase_difference,
                                                       affine=self.affine)
                    nib.save(phase_diff_nifti, base_path +
                             '_phase_difference.nii.gz')
        else:
            raise ValueError('No NIFTI file saved. The variable "maps" '
                             'should be "all" or a list of maps from '
                             '"["b0", "mask", "phase0", "phase1", '
                             '"phase_difference"]".')

        return



import os
import nibabel as nib
import numpy as np
import concurrent.futures
from tqdm import tqdm
from scipy.optimize import curve_fit


class T2:
    """
    Attributes
    ----------
    t2_map : np.ndarray
        The estimated T2 values in ms
    t2_err : np.ndarray
        The certainty in the fit of `t2` in ms
    m0_map : np.ndarray
        The estimated M0 values
    m0_err : np.ndarray
        The certainty in the fit of `m0`
    shape : tuple
        The shape of the T2 map
    n_te : int
        The number of TE used to calculate the map
    n_vox : int
        The number of voxels in the map i.e. the product of all dimensions
        apart from TE
    """

    def __init__(self, pixel_array, echo_list, affine, mask=None,
                 noise_threshold=0, method='2p_exp', multithread='auto'):
        """Initialise a T2 class instance.

        Parameters
        ----------
        pixel_array : np.ndarray
            An array containing the signal from each voxel at each echo
            time with the last dimension being time i.e. the array needed to
            generate a 3D T2 map would have dimensions [x, y, z, TE].
        echo_list : list()
            An array of the echo times used for the last dimension of the
            raw data. In milliseconds.
        affine : np.ndarray
            A matrix giving the relationship between voxel coordinates and
            world coordinates.
        mask : np.ndarray, optional
            A boolean mask of the voxels to fit. Should be the shape of the
            desired T2 map rather than the raw data i.e. omit the time
            dimension.
        noise_threshold : float, optional
            Default 0
            Voxels with magnitude less than this threshold will not be used
            when fitting. This can be useful if the noise floor of the data
            is known.
        method : {'2p_exp', '3p_exp'}, optional
            Default `2p_exp`
            The model the data is fit to. 2p_exp uses a two parameter
            exponential model (S = S0 * exp(-t / T2)) whereas 3p_exp uses a
            three parameter exponential model (S = S0 * exp(-t / T2) + b) to
            fit for noise/very long T2 components of the signal.
        multithread : bool or 'auto', optional
            Default 'auto'.
            If True, fitting will be distributed over all cores available on
            the node. If False, fitting will be carried out on a single thread.
            'auto' attempts to apply multithreading where appropriate based
            on the number of voxels being fit.
        """
        # Some sanity checks
        assert (pixel_array.shape[-1]
                == len(echo_list)), 'Number of echoes does not match the ' \
                                    'number of time frames on the last axis ' \
                                    'of pixel_array'
        assert multithread is True \
            or multithread is False \
            or multithread == 'auto', 'multithreaded must be True, ' \
                                      'False or auto. You entered {}' \
            .format(multithread)
        if method != '2p_exp' and method != '3p_exp':
            raise ValueError('method can be 2p_exp or 3p_exp only. You '
                             'specified {}'.format(method))

        self.pixel_array = pixel_array
        self.shape = pixel_array.shape[:-1]
        self.n_te = pixel_array.shape[-1]
        self.n_vox = np.prod(self.shape)
        self.affine = affine
        # Generate a mask if there isn't one specified
        if mask is None:
            self.mask = np.ones(self.shape, dtype=bool)
        else:
            self.mask = mask
            # Don't process any nan values
        self.mask[np.isnan(np.sum(pixel_array, axis=-1))] = False
        self.noise_threshold = noise_threshold
        self.method = method
        self.echo_list = echo_list
        # Auto multithreading conditions
        if multithread == 'auto':
            if self.n_vox > 20:
                multithread = True
            else:
                multithread = False
        self.multithread = multithread

        # Fit data
        if self.method == '2p_exp':
            self.t2_map, self.t2_err, \
                self.m0_map, self.m0_err \
                = self.__fit__()
        elif self.method == '3p_exp':
            self.t2_map, self.t2_err, \
                self.m0_map, self.m0_err, \
                self.b_map, self.b_err \
                = self.__fit__()

    def __fit__(self):

        # Initialise maps
        t2_map = np.zeros(self.n_vox)
        t2_err = np.zeros(self.n_vox)
        m0_map = np.zeros(self.n_vox)
        m0_err = np.zeros(self.n_vox)
        b_map = np.zeros(self.n_vox)
        b_err = np.zeros(self.n_vox)
        mask = self.mask.flatten()
        signal = self.pixel_array.reshape(-1, self.n_te)
        # Get indices of voxels to process
        idx = np.argwhere(mask).squeeze()

        # Multithreaded method
        if self.multithread:
            with concurrent.futures.ProcessPoolExecutor() as pool:
                with tqdm(total=idx.size) as progress:
                    futures = []

                    for ind in idx:
                        signal_thresh = signal[ind, :][
                            signal[ind, :] > self.noise_threshold]
                        echo_list_thresh = self.echo_list[
                            signal[ind, :] > self.noise_threshold]
                        future = pool.submit(self.__fit_signal__,
                                             signal_thresh,
                                             echo_list_thresh)
                        future.add_done_callback(lambda p: progress.update())
                        futures.append(future)

                    results = []
                    for future in futures:
                        result = future.result()
                        results.append(result)

            if self.method == '2p_exp':
                t2_map[idx], t2_err[idx], m0_map[idx], m0_err[idx] = [np.array(
                    row) for row in zip(*results)]
            elif self.method == '3p_exp':
                t2_map[idx], t2_err[idx], \
                    m0_map[idx], m0_err[idx], \
                    b_map[idx], b_err[idx] = \
                    [np.array(row) for row in zip(*results)]

        # Single threaded method
        else:
            with tqdm(total=idx.size) as progress:
                for ind in idx:
                    signal_thresh = signal[ind, :][
                        signal[ind, :] > self.noise_threshold]
                    echo_list_thresh = self.echo_list[
                        signal[ind, :] > self.noise_threshold]
                    if self.method == '2p_exp':
                        t2_map[ind], t2_err[ind], \
                            m0_map[ind], m0_err[ind] \
                            = self.__fit_signal__(signal_thresh,
                                                  echo_list_thresh)
                    elif self.method == '3p_exp':
                        t2_map[ind], t2_err[ind], \
                            m0_map[ind], m0_err[ind], \
                            b_map[ind], b_err[ind] \
                            = self.__fit_signal__(signal_thresh,
                                                  echo_list_thresh)
                    progress.update(1)

        # Reshape results to raw data shape
        t2_map = t2_map.reshape(self.shape)
        t2_err = t2_err.reshape(self.shape)
        m0_map = m0_map.reshape(self.shape)
        m0_err = m0_err.reshape(self.shape)

        if self.method == '2p_exp':
            return t2_map, t2_err, m0_map, m0_err
        elif self.method == '3p_exp':
            b_map = b_map.reshape(self.shape)
            b_err = b_err.reshape(self.shape)
            return t2_map, t2_err, m0_map, m0_err, b_map, b_err

    def __fit_signal__(self, sig, te):

        # Initialise parameters
        if self.method == '2p_exp':
            eq = two_param_eq
            bounds = ([0, 0], [1000, 1000000])
            initial_guess = [20, 10000]
        elif self.method == '3p_exp':
            eq = three_param_eq
            bounds = ([0, 0, 0], [1000, 1000000, 1000000])
            initial_guess = [20, 10000, 500]

        # Fit data to equation
        try:
            popt, pcov = curve_fit(eq, te, sig, p0=initial_guess,
                                   bounds=bounds)
        except (RuntimeError, ValueError):
            popt = np.zeros(3)
            pcov = np.zeros((3, 3))

        # Extract fits and errors from result variables
        if self.method == '2p_exp':
            if popt[0] < bounds[1][0] - 1:
                t2 = popt[0]
                m0 = popt[1]
                err = np.sqrt(np.diag(pcov))
                t2_err = err[0]
                m0_err = err[1]
            else:
                t2, m0, t2_err, m0_err = 0, 0, 0, 0

            return t2, t2_err, m0, m0_err

        elif self.method == '3p_exp':
            if popt[0] < bounds[1][0] - 1:
                t2 = popt[0]
                m0 = popt[1]
                b = popt[2]
                err = np.sqrt(np.diag(pcov))
                t2_err = err[0]
                m0_err = err[1]
                b_err = err[2]
            else:
                t2, m0, t2_err, m0_err, b, b_err = 0, 0, 0, 0, 0, 0

            return t2, t2_err, m0, m0_err, b, b_err

    def r2_map(self):
        """
        Generates the R2 map from the T2 map output by initialising this
        class.

        Parameters
        ----------
        See class attributes in __init__

        Returns
        -------
        r2 : np.ndarray
            An array containing the R2 map generated
            by the function with R2 measured in ms.
        """
        return np.reciprocal(self.t2_map)

    def to_nifti(self, output_directory=os.getcwd(), base_file_name='Output',
                 maps='all'):
        """Exports some of the T2 class attributes to NIFTI.
                        
        Parameters
        ----------
        output_directory : string, optional
            Path to the folder where the NIFTI files will be saved.
        base_file_name : string, optional
            Filename of the resulting NIFTI. This code appends the extension.
            Eg., base_file_name = 'Output' will result in 'Output.nii.gz'.
        maps : list or 'all', optional
            List of maps to save to NIFTI. This should either the string "all"
            or a list of maps from ["t2", "t2_err", "m0", "m0_err",
            "r2", "mask"].
        """
        os.makedirs(output_directory, exist_ok=True)
        base_path = os.path.join(output_directory, base_file_name)
        if maps == 'all' or maps == ['all']:
            maps = ['t2', 't2_err', 'm0', 'm0_err', 'r2', 'mask']
        if isinstance(maps, list):
            for result in maps:
                if result == 't2' or result == 't2_map':
                    t2_nifti = nib.Nifti1Image(self.t2_map, affine=self.affine)
                    nib.save(t2_nifti, base_path + '_t2_map.nii.gz')
                elif result == 't2_err':
                    t2_err_nifti = nib.Nifti1Image(self.t2_err,
                                                   affine=self.affine)
                    nib.save(t2_err_nifti, base_path + '_t2_err.nii.gz')
                elif result == 'm0' or result == 'm0_map':
                    m0_nifti = nib.Nifti1Image(self.m0_map, affine=self.affine)
                    nib.save(m0_nifti, base_path + '_m0_map.nii.gz')
                elif result == 'm0_err':
                    m0_err_nifti = nib.Nifti1Image(self.m0_err,
                                                   affine=self.affine)
                    nib.save(m0_err_nifti, base_path + '_m0_err.nii.gz')
                elif result == 'r2' or result == 'r2_map':
                    r2_nifti = nib.Nifti1Image(T2.r2_map(self),
                                               affine=self.affine)
                    nib.save(r2_nifti, base_path + '_r2_map.nii.gz')
                elif result == 'mask':
                    mask_nifti = nib.Nifti1Image(self.mask.astype(int),
                                                 affine=self.affine)
                    nib.save(mask_nifti, base_path + '_mask.nii.gz')
        else:
            raise ValueError('No NIFTI file saved. The variable "maps" '
                             'should be "all" or a list of maps from '
                             '"["t2", "t2_err", "m0", "m0_err", "r2", '
                             '"mask"]".')

        return


def two_param_eq(t, t2, m0):
    """
        Calculate the expected signal from the equation
        signal = M0 * exp(-t / T2)

        Parameters
        ----------
        t: list
            The times the signal will be calculated at
        t2: float
            The T2 of the signal
        m0: float
            The M0 of the signal

        Returns
        -------
        signal: np.ndarray
            The expected signal
        """
    return np.sqrt(np.square(m0 * np.exp(-t / t2)))


def three_param_eq(t, t2, m0, b):
    """
        Calculate the expected signal from the equation
        signal = M0 * exp(-t / T2) + b

        Parameters
        ----------
        t: list
            The times the signal will be calculated at
        t2: float
            The T2 of the signal
        m0: float
            The M0 of the signal
        b: float
            The baseline noise floor of the signal

        Returns
        -------
        signal: np.ndarray
            The expected signal
        """
    return np.sqrt(np.square(m0 * np.exp(-t / t2) + b))



import csv
import os
import nibabel as nib
import numpy as np

from segment import Tkv


class Segmentation(nib.Nifti1Image):
    """
    Attributes
    ----------
    volumes : dict
        Total, left and right kidney volumes in milliliters.
    """

    def __init__(self, pixel_array, affine, post_process=True, binary=True,
                 weights=None):
        """Initialise a whole kidney segmentation class instance.

        Parameters
        ----------
        pixel_array : np.ndarray
            Array containing a T2-weighted FSE image.
        affine : np.ndarray
            A matrix giving the relationship between voxel coordinates and
            world coordinates.
        post_process : bool, optional
            Default True
            Keep only the two largest connected volumes in the mask. Note
            this may cause issue with subjects that have more or less than
            two kidneys.
        binary : bool, optional
            Default True.
            If True, the mask returned will be an array of ints, where 1
            represents voxels that are renal tissue and 0 represents
            voxels that are not renal tissue. If False, the mask returned
            will be the probability that each voxel is renal tissue.
        weights : str, optional
            Default 'none'.
            The path to custom neural network weights. If 'none',
            the default, all-purpose, weights will be downloaded from the
            internet and used.
        """

        super().__init__(pixel_array, affine)
        self.pixel_array = pixel_array
        self._nifti = nib.Nifti1Image(self.pixel_array, self.affine)
        self._seg_obj = Tkv(self._nifti)
        self._mask = self._seg_obj.get_mask(post_process=post_process,
                                            binary=binary,
                                            weights_path=weights)
        self._kidneys = (self._mask > 0.5) * 1
        self._kidneys[:self.shape[0]//2] *= 2
        self.volumes = {'total': self._calculate_volume(self._mask > 0.5),
                        'left': self._calculate_volume(self._kidneys == 1),
                        'right': self._calculate_volume(self._kidneys == 2)}

    def get_mask(self):
        """
        Returns the estimated mask from the provided input data.

        Returns
        -------
        mask : np.ndarray
            The estimated mask.
        """
        return self._mask

    def get_kidneys(self):
        """
        Returns a mask where 0 represents voxels that are not renal tissue,
        1 represents voxels that are the left kidney and 2 represents voxels
        that are the right kidney.

        Returns
        -------
        mask : np.ndarray
            Mask with each kidney represented by a different int.
        """
        return self._kidneys

    def get_left_kidney(self):
        """
        Returns a binary mask where True is the left kidney.

        Returns
        -------
        left_kidney : np.ndarray
            Binary mask of left kidney.
        """
        left_kidney = (self._kidneys == 1) * 1
        return left_kidney

    def get_right_kidney(self):
        """
        Returns a binary mask where True is the right kidney.

        Returns
        -------
        right_kidney : np.ndarray
            Binary mask of right kidney.
        """
        right_kidney = (self._kidneys == 2) * 1
        return right_kidney

    def get_volumes(self):
        """
        Return the volume of each kidney and total kidney volume in
        milliliters.

        Returns
        -------
        volumes : dict
            Total, left and right kidney volumes in milliliters.
        """
        return self.volumes

    def save_volumes_csv(self, path):
        """
        Save the total, left and right kidney volumes to a csv file.

        Parameters
        ----------
        path : str
            Path to the desired csv file.
        """
        with open(path, 'w', newline='') as f:
            w = csv.DictWriter(f, self.volumes.keys())
            w.writeheader()
            w.writerow(self.volumes)

    def get_tkv(self):
        """
        Get the total kidney volume in milliliters.
        Returns
        -------
        tkv : float
            Total kidney volume
        """
        return self.volumes['total']

    def get_lkv(self):
        """
        Get the left kidney volume in milliliters.
        Returns
        -------
        lkv : float
            Left kidney volume
        """
        return self.volumes['left']

    def get_rkv(self):
        """
        Get the right kidney volume in milliliters.
        Returns
        -------
        rkv : float
            Right kidney volume
        """
        return self.volumes['right']

    def to_nifti(self, output_directory=os.getcwd(), base_file_name='Output',
                 maps='all'):
        """Exports masks to NIFTI.

        Parameters
        ----------
        output_directory : string, optional
            Path to the folder where the NIFTI files will be saved.
        base_file_name : string, optional
            Filename of the resulting NIFTI. This code appends the extension.
            Eg., base_file_name = 'Output' will result in 'Output.nii.gz'.
        maps : list or 'all', optional
            List of maps to save to NIFTI. This should either the string "all"
            or a list of maps from ["mask", "left", "right", "individual"]
        """
        os.makedirs(output_directory, exist_ok=True)
        base_path = os.path.join(output_directory, base_file_name)
        if maps == 'all' or maps == ['all']:
            maps = ['mask', 'left', 'right', 'individual']

        if isinstance(maps, list):
            for result in maps:
                if result == 'mask':
                    mask_nifti = nib.Nifti1Image(self._mask, self.affine)
                    nib.save(mask_nifti, base_path + '_mask.nii.gz')
                elif result == 'left':
                    left_nifti = nib.Nifti1Image(self.get_left_kidney(),
                                                 self.affine)
                    nib.save(left_nifti, base_path + '_left_kidney.nii.gz')
                elif result == 'right':
                    right_nifti = nib.Nifti1Image(self.get_right_kidney(),
                                                  self.affine)
                    nib.save(right_nifti, base_path + '_right_kidney.nii.gz')
                elif result == 'individual':
                    ind_nifti = nib.Nifti1Image(self.get_kidneys(),
                                                self.affine)
                    nib.save(ind_nifti, base_path +
                             '_individual_kidneys.nii.gz')
        else:
            raise ValueError('No NIFTI file saved. The variable "maps" '
                             'should be "all" or a list of maps from '
                             '"["mask", "left", "right", "individual"]".')

    def _calculate_volume(self, mask):
        mask = mask > 0.5
        return np.sum(mask) * np.prod(self.header.get_zooms()) / 1000



"""
Diffusion imaging module

"""
import os
import nibabel as nib
import numpy as np
import warnings

from dipy.core.gradients import gradient_table, unique_bvals_tolerance
from dipy.reconst.dti import TensorModel
from tqdm import tqdm


def make_gradient_scheme(bvals, bvecs, normalize=True, one_bzero=True):
    """Make gradient scheme from list of bvals and bvecs

    Parameters
    ----------
    bvals : list
        b-values in s/mm2
    bvecs : list of lists
        bvectors (e.g. [[0, 0, 1], [1, 0, 0]])
    normalize : bool, optional (default True)
        Rescales bvecs to have unit length
    one_bzero : bool, optional (default True)
        Ensures gradient scheme only includes one b=0 measurement
        If this is True and bvals does not contain any b=0, a b=0 measurement
        will be included at the start of the acquisition

    Returns
    -------
    string
        gradient scheme with one line per measurement/volume as follows:
        bvec1_x   bvec1_y   bvec1_z   bval1
        bvec2_x   bvec2_y   bvec2_z   bval2
        ...
        bvecN_x   bvecN_y   bvecN_z   bvalN

    Notes
    -----
    This function was created to generate a diffusion scheme for the UKRIN-MAPS
    multishell acquisition where all the nonzero b-values have the same number
    of directions. Currently this does not provide features to generate
    schemes where different shells have different numbers of directions.
    This gradient scheme format was decided with the following in mind:
        1) can be easily written to a text file to allow modifications to it
           to be done without coding
        2) not vendor specific
        3) could be useful as a starting point to convert these schemes to
           vendor-specific formats

    """
    if 0 not in bvals and one_bzero:
        bvals.insert(0, 0)

    if normalize:
        # Rescale bvecs to have norm 1
        bvecs = [v/np.linalg.norm(v) for v in bvecs]

    bvecs = [np.round(x, 8) for x in bvecs]

    # Make gradient scheme
    bzero_counter = 0
    gradient_scheme = ""
    for bvec in bvecs:
        for bval in bvals:
            if bval == 0 and one_bzero and bzero_counter > 0:
                continue
            else:
                gradient_scheme = (f"{gradient_scheme}"
                                   f"{str(bvec[0]).rjust(11)}  "
                                   f"{str(bvec[1]).rjust(11)}  "
                                   f"{str(bvec[2]).rjust(11)}  "
                                   f"{str(bval).rjust(5)}\n")
                if bval == 0:
                    bzero_counter += 1

    # Remove last newline
    if gradient_scheme[-1] == '\n':
        gradient_scheme = gradient_scheme[:-1:]

    return gradient_scheme


class ADC:
    """
    Attributes
    ----------
    adc : np.ndarray
        The estimated ADC in mm^2/s.
    adc_err : np.ndarray
        The certainty in the fit of `adc` in mm^2/s.
    shape : tuple
        The shape of the ADC map.
    n_vox : int
        The number of voxels in the map.
    bvals : 1d numpy array
        All b-values that will be used to generate the maps in s/mm^2.
    u_bvals : 1d numpy array
        The unique b-values used in the experiment e.g. if the experiment
        acquires a single b-0 volume and 64 volumes with b=600 s/mm^2 in
        different directions, u_bvals will be [0, 600].
    n_bvals : int
        The number of unique b-values acquired in the experiment.
    n_grad : int
        Total number of diffusion values/vectors acquired e.g. if the
        experiment acquires six directions at 10 gradient strengths and a
        b-0 volume, n_grad will be 61.
    pixel_array_mean : np.ndarray
        The average of the `pixel_array`across bvecs e.g. if `pixel_array`
        contains six volumes acquired with b=600 s/mm^2 in different
        directions, these six volumes will be averaged together.
    """
    def __init__(self, pixel_array, affine, bvals, mask=None, ukrin_b=False):
        """Initialise a ADC class instance.

        Parameters
        ----------
        pixel_array : (..., N) np.ndarray
            A array containing the signal from each voxel at each
            diffusion sensitising parameter. The final dimension should be
            different diffusion weightings/directions.
        affine : np.ndarray
            A matrix giving the relationship between voxel coordinates and
            world coordinates.
        bvals : (N,) np.array
            An array of the b-values used for the last dimension of the raw
            data. In s/mm^2.
        mask : np.ndarray, optional
            A boolean mask of the voxels to fit. Should be the shape of the
            desired map rather than the raw data i.e. omit the last dimension.
        ukrin_b : bool, optional
            If True, only b-values of 0, 100, 200 and 800 s/mm^2 will be
            included in the ADC fit. This aligns with Ljimani A, et al.
            Consensus-based technical recommendations for clinical translation
            of renal diffusion-weighted MRI.
            Magn Reson Mater Phy 2020;33:177–195
            doi: 10.1007/s10334-019-00790-y.
            If False, all b-values supplied will be used to fit ADC.
        """
        ukrin_b_test = np.array([0, 100, 200, 800])
        # Sanity checks
        assert (pixel_array.shape[-1]
                == len(bvals)), 'Number of bvals does not match number of ' \
                                'gradients in pixel_array'

        if ukrin_b:
            self.b_mask = np.isin(bvals, ukrin_b_test)
        else:
            self.b_mask = np.full(len(bvals), True, dtype=bool)

        self.pixel_array = pixel_array[..., self.b_mask]
        self.shape = pixel_array.shape[:-1]
        self.n_vox = np.prod(self.shape)
        self.bvals = bvals[self.b_mask]
        self.n_grad = len(self.bvals)
        self.u_bvals = unique_bvals_tolerance(self.bvals, 1)
        self.n_bvals = len(self.u_bvals)
        self.affine = affine
        # Generate a mask if there isn't one specified
        if mask is None:
            self.mask = np.ones(self.shape, dtype=bool)
        else:
            self.mask = mask
            # Don't process any nan values
        self.mask[np.isnan(np.sum(pixel_array, axis=-1))] = False
        self.mask[np.sum(pixel_array <= 0, axis=-1, dtype=bool)] = False
        self.pixel_array = np.nan_to_num(self.pixel_array)

        self.pixel_array_mean = self.__mean_over_directions__()

        self.adc, self.adc_err = self.__fit__()

    def __mean_over_directions__(self):
        """
        Calculates the mean signal across different directions at each unique
        b-value e.g. if `pixel_array` contains six volumes acquired with
        b=600 s/mm^2 in different directions, these six volumes will be
        averaged together.

        Returns
        -------
        pixel_array_mean : np.ndarray
            The average of the `pixel_array` across bvecs
        """
        pixel_array_mean = np.zeros((*self.shape, self.n_bvals))
        for ind, bval in enumerate(self.u_bvals):
            pixel_array_mean[..., ind]\
                = np.mean(self.pixel_array[..., self.bvals == bval], axis=-1)
        return pixel_array_mean

    def __fit__(self):
        # Initialise maps
        adc_map = np.zeros(self.n_vox)
        adc_err = np.zeros(self.n_vox)

        mask = self.mask.flatten()
        signal = self.pixel_array_mean.reshape(-1, self.n_bvals)
        idx = np.argwhere(mask).squeeze()
        with tqdm(total=idx.size) as progress:
            for ind in idx:
                sig = signal[ind, :]
                adc_map[ind], adc_err[ind] = \
                    self.__fit_signal__(sig, self.u_bvals)
                progress.update(1)
        adc_map[adc_map < 0] = 0
        adc_err[adc_map < 0] = 0

        # Reshape results into raw data shape
        adc_map = adc_map.reshape(self.shape)
        adc_err = adc_err.reshape(self.shape)

        return adc_map, adc_err

    @staticmethod
    def __fit_signal__(sig, bvals):
        try:
            popt, pvar = np.polyfit(bvals[sig > 0], np.log(sig[sig > 0]), 1,
                                    cov=True)
            adc = -popt[0]
            adc_err = np.sqrt(pvar[0, 0])
        except np.linalg.LinAlgError:
            adc = 0
            adc_err = 0

        return adc, adc_err

    def to_nifti(self, output_directory=os.getcwd(), base_file_name='Output',
                 maps='all'):
        """Exports maps generated by the ADC class as NIFTI.

        Parameters
        ----------
        output_directory : string, optional
            Path to the folder where the NIFTI files will be saved.
        base_file_name : string, optional
            Filename of the resulting NIFTI. This code appends the extension.
            Eg., base_file_name = 'Output' will result in 'Output.nii.gz'.
        maps : list or 'all', optional
            List of maps to save to NIFTI. This should either the string "all"
            or a list of maps from ["adc", "adc_err", "mask"].
        """
        os.makedirs(output_directory, exist_ok=True)
        base_path = os.path.join(output_directory, base_file_name)
        if maps == 'all' or maps == ['all']:
            maps = ['adc', 'adc_err', 'mask']
        if isinstance(maps, list):
            for result in maps:
                if result == 'adc' or result == 'adc_map':
                    adc_nifti = nib.Nifti1Image(self.adc, affine=self.affine)
                    nib.save(adc_nifti, base_path + '_adc_map.nii.gz')
                elif result == 'adc_err' or result == 'adc_err_map':
                    adc_err_nifti = nib.Nifti1Image(self.adc_err,
                                                    affine=self.affine)
                    nib.save(adc_err_nifti, base_path + '_adc_err.nii.gz')
                elif result == 'mask':
                    mask_nifti = nib.Nifti1Image(self.mask.astype(int),
                                                 affine=self.affine)
                    nib.save(mask_nifti, base_path + '_mask.nii.gz')
        else:
            raise ValueError('No NIFTI file saved. The variable "maps" '
                             'should be "all" or a list of maps from '
                             '"["adc", "adc_err", "mask"]".')


class DTI:
    """
    Attributes
    ----------
    md : np.ndarray
        The estimated mean diffusivity values in mm^2/s.
    fa : np.ndarray
        The estimated fractional anisotropy values.
    color_fa : np.ndarray
        The estimated directional fractional anisotropy represented as red,
        green and blue corresponding to correspond to fractional anisotropy
        in the x, y and z directions respectively.
    shape : tuple
        The shape of the resulting maps
    bvals : 1d numpy array
        All b-values that will be used to generate the maps in s/mm^2.
    u_bvals : 1d numpy array
        The unique b-values used in the experiment e.g. if the experiment
        acquires a single b-0 volume and 64 volumes with b=600 s/mm^2 in
        different directions, u_bvals will be [0, 600].
    n_bvals : int
        The number of unique b-values acquired in the experiment.
    bvecs : (N, 3) numpy array
        All b-vectors that will be used to generate the maps.
    u_bvecs : (M, 3) numpy array
        The unique b-vectors used in the experiment e.g. if the experiment
        acquires six directions at 10 gradient strengths, u_bvecs will be a
        6 x 3 numpy array.
    n_bvecs : int
        The number of unique b-vectors acquired in the experiment.
    n_grad : int
        Total number of diffusion values/vectors acquired e.g. if the
        experiment acquires six directions at 10 gradient strengths and a
        b-0 volume, n_grad will be 61.
    gtab : dipy GradientTable
        The dipy gradient table used to generate maps.
    tensor_fit : dipy TensorModel after fitting
        The fit dipy tensor model, can be used to recall additional parameters.
    """
    def __init__(self, pixel_array, affine, bvals, bvecs, mask=None,
                 ukrin_b=False):
        """Initialise a DTI class instance.

        Parameters
        ----------
        pixel_array : (..., N) np.ndarray
            A array containing the signal from each voxel at each
            diffusion sensitising parameter. The final dimension should be
            different diffusion weightings/directions.
        affine : np.ndarray
            A matrix giving the relationship between voxel coordinates and
            world coordinates.
        bvals : (N,) np.array
            An array of the b-values used for the last dimension of the raw
            data. In s/mm^2.
        bvecs : (N, 3) np.array
            An array of the b-vectors used for the last dimension of the raw
            data. In s/mm^2.
        mask : np.ndarray, optional
            A boolean mask of the voxels to fit. Should be the shape of the
            desired map rather than the raw data i.e. omit the last dimension.
        ukrin_b : bool, optional
            If True, only b-values of 0, 100, 200 and 800 s/mm^2 will be
            included in the fit. This aligns with Ljimani A, et al.
            Consensus-based technical recommendations for clinical translation
            of renal diffusion-weighted MRI.
            Magn Reson Mater Phy 2020;33:177–195
            doi: 10.1007/s10334-019-00790-y.
            If False, all b-values supplied will be used.
        """
        ukrin_b_test = np.array([0, 100, 200, 800])
        # Some sanity checks
        assert (pixel_array.shape[-1]
                == len(bvals)), 'Number of bvals does not match number of ' \
                                'gradients in pixel_array'
        if bvecs.shape[1] != 3 and bvecs.shape[0] == 3:
            bvecs = bvecs.T
            warnings.warn(f'bvecs should be (N, 3). Because your bvecs array '
                          'is {bvecs.shape} it has been transposed to {'
                          'bvecs.T.shape}.')
        assert (bvecs.shape[1] == 3)
        assert (pixel_array.shape[-1] == bvecs.shape[0]), 'Number of bvecs ' \
                                                          'does not match ' \
                                                          'number of ' \
                                                          'gradients in ' \
                                                          'pixel_array'
        if ukrin_b:
            self.b_mask = np.isin(bvals, ukrin_b_test)
        else:
            self.b_mask = np.full(len(bvals), True, dtype=bool)

        self.pixel_array = pixel_array[..., self.b_mask]
        self.shape = self.pixel_array.shape[:-1]
        self.bvals = bvals[self.b_mask]
        self.bvecs = bvecs[self.b_mask, :]
        self.n_grad = len(self.bvals)
        self.u_bvals = unique_bvals_tolerance(self.bvals, 1)
        self.n_bvals = len(self.u_bvals)
        self.u_bvecs = np.unique(self.bvecs, axis=0)
        self.n_bvecs = len(self.u_bvecs)
        self.affine = affine
        self.mask = mask
        self.gtab = gradient_table(self.bvals, self.bvecs, b0_threshold=0)
        tensor_model = TensorModel(self.gtab)
        self.tensor_fit = tensor_model.fit(self.pixel_array, mask=self.mask)
        self.md = self.tensor_fit.md
        self.fa = self.tensor_fit.fa
        self.color_fa = self.tensor_fit.color_fa

    def to_nifti(self, output_directory=os.getcwd(), base_file_name='Output',
                 maps='all'):
        """Exports maps generated by the DTI class as NIFTI.

        Parameters
        ----------
        output_directory : string, optional
            Path to the folder where the NIFTI files will be saved.
        base_file_name : string, optional
            Filename of the resulting NIFTI. This code appends the extension.
            Eg., base_file_name = 'Output' will result in 'Output.nii.gz'.
        maps : list or 'all', optional
            List of maps to save to NIFTI. This should either the string "all"
            or a list of maps from ["md", "fa", "color_fa", "mask"].
        """
        os.makedirs(output_directory, exist_ok=True)
        base_path = os.path.join(output_directory, base_file_name)
        if maps == 'all' or maps == ['all']:
            maps = ['md', 'fa', 'color_fa', 'mask']
        if isinstance(maps, list):
            for result in maps:
                if result == 'md' or result == 'md_map':
                    md_nifti = nib.Nifti1Image(self.md, affine=self.affine)
                    nib.save(md_nifti, base_path + '_md_map.nii.gz')
                elif result == 'fa' or result == 'fa_map':
                    fa_nifti = nib.Nifti1Image(self.fa, affine=self.affine)
                    nib.save(fa_nifti, base_path + '_fa_map.nii.gz')
                elif result == 'color_fa' or result == 'color_fa_map':
                    color_fa_nifti = nib.Nifti1Image(self.color_fa,
                                                     affine=self.affine)
                    nib.save(color_fa_nifti, base_path +
                             '_color_fa_map.nii.gz')
                elif result == 'mask':
                    mask_nifti = nib.Nifti1Image(self.mask.astype(int),
                                                 affine=self.affine)
                    nib.save(mask_nifti, base_path + '_mask.nii.gz')
        else:
            raise ValueError('No NIFTI file saved. The variable "maps" '
                             'should be "all" or a list of maps from '
                             '"["md", "fa", "color_fa", "mask"]".')



import os
import shutil
import numpy as np
import numpy.testing as npt
import pandas as pd
import pytest
from ukat.data import fetch
from ukat.vessels.phase_contrast import PhaseContrast, convert_to_velocity
from ukat.utils import arraystats


class TestPC:
    # Create arrays for testing
    correct_signal = np.arange(2000).reshape((10, 10, 20))  # velocity array
    two_dim_signal = np.arange(100).reshape((10, 10))
    vel_encoding = 100
    affine = np.eye(4)
    mask = np.ones(correct_signal.shape, dtype=bool)

    # Gold standard: [mean, std, min, max] of velocity_array when input
    # is `correct_signal`
    gold_standard = [999.5, 577.3501970208376, 0, 1999]

    # Expected PhaseContrast statistics for ROI, velocity and flow
    num_pixels_standard = [100] * 20
    area_standard = [1.0] * 20  # Each pixel is 1mm*1mm, but area is in cm2
    min_vel_standard = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13,
                        14, 15, 16, 17, 18, 19]
    mean_vel_standard = [990.0, 991.0, 992.0, 993.0, 994.0, 995.0, 996.0,
                         997.0, 998.0, 999.0, 1000.0, 1001.0, 1002.0,
                         1003.0, 1004.0, 1005.0, 1006.0, 1007.0, 1008.0,
                         1009.0]
    max_vel_standard = [1980, 1981, 1982, 1983, 1984, 1985, 1986, 1987,
                        1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995,
                        1996, 1997, 1998, 1999]
    std_vel_standard = [577.3214009544423] * 20
    rbf_standard = [59400.0, 59460.0, 59520.0, 59580.0, 59640.0, 59700.0,
                    59760.0, 59820.0, 59880.0, 59940.0, 60000.0, 60060.0,
                    60120.0, 60180.0, 60240.0, 60300.0, 60360.0, 60420.0,
                    60480.0, 60540.0]
    mean_vel_global_standard = 999.5
    rbf_mean_standard = 59970.0
    resistive_index_standard = 0.018830525272547076

    # Velocity array gold standard stats assuming the `correct_signal`
    # is a phase image.
    artificial_vel_array_standard = [0.0, 181.47064907546604,
                                     -314.1592653589793, 314.1592653589793]

    def test_velocity_array(self):
        vel_array_whole_mask = PhaseContrast(self.correct_signal,
                                             self.affine,
                                             self.mask).velocity_array
        vel_stats = arraystats.ArrayStats(vel_array_whole_mask).calculate()
        npt.assert_allclose([vel_stats["mean"]["3D"], vel_stats["std"]["3D"],
                            vel_stats["min"]["3D"], vel_stats["max"]["3D"]],
                            self.gold_standard, rtol=1e-7, atol=1e-9)

    def test_input_errors(self):
        # Check that it fails when input pixel_array has incorrect shape
        with pytest.raises(ValueError):
            PhaseContrast(self.two_dim_signal, self.affine, self.mask)

    def test_affine_pixel_spacing(self):
        pc_obj = PhaseContrast(self.correct_signal, self.affine, self.mask)
        npt.assert_allclose(pc_obj.pixel_spacing, [1.0, 1.0],
                            rtol=1e-7, atol=1e-9)
        affine_test_1 = np.eye(4)
        affine_test_1[:, 0] = [1, 2, 2, 0]
        pc_obj = PhaseContrast(self.correct_signal, affine_test_1, self.mask)
        npt.assert_allclose(pc_obj.pixel_spacing, [1.0, 3.0],
                            rtol=1e-7, atol=1e-9)
        affine_test_2 = affine_test_1
        affine_test_2[:, 1] = [2, 1, 2, 0]
        pc_obj = PhaseContrast(self.correct_signal, affine_test_2, self.mask)
        npt.assert_allclose(pc_obj.pixel_spacing, [3.0, 3.0],
                            rtol=1e-7, atol=1e-9)

    def test_mask(self):
        # Create a mask where the first 5 rows of each phase
        # are False and the rest is True.
        partial_mask = np.ones(self.correct_signal.shape, dtype=bool)
        partial_mask[:5, ...] = False

        all_pixels = PhaseContrast(self.correct_signal, self.affine, self.mask)
        masked_pixels = PhaseContrast(self.correct_signal,
                                      self.affine, partial_mask)

        assert (all_pixels.velocity_array !=
                masked_pixels.velocity_array).any()
        assert (all_pixels.mean_velocity !=
                masked_pixels.mean_velocity).any()
        assert (all_pixels.mean_velocity_global !=
                masked_pixels.mean_velocity_global)
        assert (all_pixels.rbf != masked_pixels.rbf).any()
        assert all_pixels.mean_rbf != masked_pixels.mean_rbf

    def test_output_parameters(self):
        pc_obj = PhaseContrast(self.correct_signal, self.affine, self.mask)
        npt.assert_allclose(pc_obj.num_pixels, self.num_pixels_standard,
                            rtol=1e-7, atol=1e-9)
        npt.assert_allclose(pc_obj.area, self.area_standard,
                            rtol=1e-7, atol=1e-9)
        npt.assert_allclose(pc_obj.min_velocity,
                            self.min_vel_standard,
                            rtol=1e-7, atol=1e-9)
        npt.assert_allclose(pc_obj.mean_velocity,
                            self.mean_vel_standard,
                            rtol=1e-7, atol=1e-9)
        npt.assert_allclose(pc_obj.max_velocity,
                            self.max_vel_standard,
                            rtol=1e-7, atol=1e-9)
        npt.assert_allclose(pc_obj.std_velocity,
                            self.std_vel_standard,
                            rtol=1e-7, atol=1e-9)
        npt.assert_allclose(pc_obj.rbf, self.rbf_standard,
                            rtol=1e-7, atol=1e-9)
        npt.assert_allclose(pc_obj.mean_velocity_global,
                            self.mean_vel_global_standard,
                            rtol=1e-7, atol=1e-9)
        npt.assert_allclose(pc_obj.mean_rbf, self.rbf_mean_standard,
                            rtol=1e-7, atol=1e-9)
        npt.assert_allclose(pc_obj.resistive_index,
                            self.resistive_index_standard,
                            rtol=1e-7, atol=1e-9)

    def test_get_stats_table(self):
        pc_obj = PhaseContrast(self.correct_signal, self.affine, self.mask)
        stats_table = pc_obj.get_stats_table()
        assert type(stats_table) == pd.DataFrame
        number_rows = len(stats_table)
        number_columns = len(stats_table.columns)
        assert number_rows == 20
        assert number_columns == 7

    def test_to_csv(self):
        os.makedirs('test_output', exist_ok=True)
        csv_path = os.path.join('test_output', "pc_test_output.csv")
        pc_obj = PhaseContrast(self.correct_signal, self.affine, self.mask)
        # Save .csv and and open it to compare if it's the same as expected.
        pc_obj.to_csv(csv_path)
        df = pd.read_csv(csv_path)
        shutil.rmtree('test_output')
        list_rows = df.values.tolist()
        # Check if the content in the variable 'list_rows' is as expected.
        assert [row[1] for row in list_rows] == self.rbf_standard
        assert [row[2] for row in list_rows] == self.area_standard
        assert [row[3] for row in list_rows] == self.num_pixels_standard
        assert [row[4] for row in list_rows] == self.mean_vel_standard
        assert [row[5] for row in list_rows] == self.min_vel_standard
        assert [row[6] for row in list_rows] == self.max_vel_standard
        assert [row[7] for row in list_rows] == self.std_vel_standard

    def test_plot(self):
        # Since we cannot see the actual plots during the testing, the best
        # approach is to check if it fails with an incorrect stat call.
        with pytest.raises(ValueError):
            pc_obj = PhaseContrast(self.correct_signal, self.affine, self.mask)
            pc_obj.plot('incorrect_stat_name')

    def test_to_nifti(self):
        # Create a PC instance and test different export to NIFTI scenarios
        pc_obj = PhaseContrast(self.correct_signal, self.affine, self.mask)
        os.makedirs('test_output', exist_ok=True)
        # Check all is saved.
        pc_obj.to_nifti(output_directory='test_output',
                        base_file_name='pctest', maps='all')
        output_files = os.listdir('test_output')
        assert len(output_files) == 2
        assert 'pctest_velocity_array.nii.gz' in output_files
        assert 'pctest_mask.nii.gz' in output_files

        for f in os.listdir('test_output'):
            os.remove(os.path.join('test_output', f))

        # Check that no files are saved.
        pc_obj.to_nifti(output_directory='test_output',
                        base_file_name='pctest', maps=[])
        output_files = os.listdir('test_output')
        assert len(output_files) == 0

        # Check that only velocity_array saved.
        pc_obj.to_nifti(output_directory='test_output',
                        base_file_name='pctest', maps=['velocity_array'])
        output_files = os.listdir('test_output')
        assert len(output_files) == 1
        assert 'pctest_velocity_array.nii.gz' in output_files

        for f in os.listdir('test_output'):
            os.remove(os.path.join('test_output', f))

        # Check that it fails when no maps are given
        with pytest.raises(ValueError):
            pc_obj = PhaseContrast(self.correct_signal, self.affine, self.mask)
            pc_obj.to_nifti(output_directory='test_output',
                            base_file_name='pctest', maps='')

        # Delete 'test_output' folder
        shutil.rmtree('test_output')

    def test_convert_to_velocity(self):
        # Assuming that correct_signal is a phase image and not velocity array.
        vel_array_calculated = convert_to_velocity(self.correct_signal,
                                                   self.vel_encoding)
        vel_stats = arraystats.ArrayStats(vel_array_calculated).calculate()
        npt.assert_allclose([vel_stats["mean"]["3D"], vel_stats["std"]["3D"],
                            vel_stats["min"]["3D"], vel_stats["max"]["3D"]],
                            self.artificial_vel_array_standard,
                            rtol=1e-7, atol=1e-9)

    def test_real_data(self):
        # Get left PC test data
        _, velocity, mask, affine, _ = fetch.phase_contrast_left_philips()

        # Gold standard statistics
        gold_standard_left = [0.009531885391710501, 0.5249643990944701,
                              0.0, 61.17216385900974]
        avrg_vel_left = 25.57730254026419
        rbf_mean_standard_left = 448.3799164652045
        resistive_index_left = 0.6089871033735511

        pc_obj = PhaseContrast(velocity, affine, mask)
        vel_stats = arraystats.ArrayStats(pc_obj.velocity_array).calculate()
        npt.assert_allclose([vel_stats["mean"]["3D"], vel_stats["std"]["3D"],
                            vel_stats["min"]["3D"], vel_stats["max"]["3D"]],
                            gold_standard_left, rtol=0.01, atol=0)
        npt.assert_allclose(pc_obj.mean_velocity_global, avrg_vel_left,
                            rtol=0.01, atol=0)
        npt.assert_allclose(pc_obj.mean_rbf, rbf_mean_standard_left,
                            rtol=0.01, atol=0)
        npt.assert_allclose(pc_obj.resistive_index, resistive_index_left,
                            rtol=0.01, atol=0)

        # Get right PC test data
        _, velocity, mask, affine, _ = fetch.phase_contrast_right_philips()

        # Gold standard statistics
        gold_standard_right = [0.014580785805088049, 0.7169850435217269,
                               0.0, 97.36264065280557]
        avrg_vel_right = 28.752678965667474
        rbf_mean_standard_right = 685.880219
        resistive_index_right = 0.6713997583484632

        pc_obj = PhaseContrast(velocity, affine, mask)
        vel_stats = arraystats.ArrayStats(pc_obj.velocity_array).calculate()
        npt.assert_allclose([vel_stats["mean"]["3D"], vel_stats["std"]["3D"],
                            vel_stats["min"]["3D"], vel_stats["max"]["3D"]],
                            gold_standard_right, rtol=0.01, atol=0)
        npt.assert_allclose(pc_obj.mean_velocity_global, avrg_vel_right,
                            rtol=0.01, atol=0)
        npt.assert_allclose(pc_obj.mean_rbf, rbf_mean_standard_right,
                            rtol=0.01, atol=0)
        npt.assert_allclose(pc_obj.resistive_index, resistive_index_right,
                            rtol=0.01, atol=0)


# Delete the NIFTI test folder recursively if any of the unit tests failed
if os.path.exists('test_output'):
    shutil.rmtree('test_output')



import os
import shutil
import numpy as np
import numpy.testing as npt
import pytest
from ukat.data import fetch
from ukat.mapping.t2star import T2Star, two_param_eq
from ukat.utils import arraystats


class TestT2Star:
    t2star = 50
    m0 = 3000
    t = np.arange(5, 39, 3)

    # The idea signal produced by the equation M8 * exp(-t / T2*) where
    # M0 = 5000 and T2* = 50 ms at 12 echo times between 5 and 38 ms
    correct_signal = np.array([2714.51225411, 2556.4313669, 2407.55639389,
                               2267.35122437, 2135.31096829, 2010.96013811,
                               1893.85093652, 1783.56164391, 1679.6950997,
                               1581.87727213, 1489.75591137, 1402.99928103])
    affine = np.eye(4)
    def test_two_param_eq(self):
        signal = two_param_eq(self.t, self.t2star, self.m0)
        npt.assert_allclose(signal, self.correct_signal, rtol=1e-6, atol=1e-8)

    def test_loglin_fit(self):
        # Make the signal into a 4D array
        signal_array = np.tile(self.correct_signal, (10, 10, 3, 1))

        # Multithread
        mapper = T2Star(signal_array, self.t, self.affine, method='loglin',
                        multithread=True)
        assert mapper.shape == signal_array.shape[:-1]
        npt.assert_almost_equal(mapper.t2star_map.mean(), self.t2star)
        assert np.isnan(mapper.t2star_err.mean())
        npt.assert_almost_equal(mapper.m0_map.mean(), self.m0)
        npt.assert_almost_equal(mapper.r2star_map().mean(), 1 / self.t2star)

        # Single Threaded
        mapper = T2Star(signal_array, self.t, self.affine, method='loglin',
                        multithread=False)
        assert mapper.shape == signal_array.shape[:-1]
        npt.assert_almost_equal(mapper.t2star_map.mean(), self.t2star)
        assert np.isnan(mapper.t2star_err.mean())
        npt.assert_almost_equal(mapper.m0_map.mean(), self.m0)
        npt.assert_almost_equal(mapper.r2star_map().mean(), 1 / self.t2star)

        # Auto Threaded
        mapper = T2Star(signal_array, self.t, self.affine, method='loglin',
                        multithread='auto')
        assert mapper.shape == signal_array.shape[:-1]
        npt.assert_almost_equal(mapper.t2star_map.mean(), self.t2star)
        assert np.isnan(mapper.t2star_err.mean())
        npt.assert_almost_equal(mapper.m0_map.mean(), self.m0)
        npt.assert_almost_equal(mapper.r2star_map().mean(), 1 / self.t2star)

    def test_2p_exp_fit(self):
        # Make the signal into a 4D array
        signal_array = np.tile(self.correct_signal, (10, 10, 3, 1))

        # Multithread
        mapper = T2Star(signal_array, self.t, self.affine, method='2p_exp',
                        multithread=True)
        assert mapper.shape == signal_array.shape[:-1]
        npt.assert_almost_equal(mapper.t2star_map.mean(), self.t2star)
        npt.assert_almost_equal(mapper.t2star_err.mean(), 7.395706644238e-11)
        npt.assert_almost_equal(mapper.m0_map.mean(), self.m0)
        npt.assert_almost_equal(mapper.r2star_map().mean(), 1 / self.t2star)

        # Single Threaded
        mapper = T2Star(signal_array, self.t, self.affine, method='2p_exp',
                        multithread=False)
        assert mapper.shape == signal_array.shape[:-1]
        npt.assert_almost_equal(mapper.t2star_map.mean(), self.t2star)
        npt.assert_almost_equal(mapper.t2star_err.mean(), 7.395706644238e-11)
        npt.assert_almost_equal(mapper.m0_map.mean(), self.m0)
        npt.assert_almost_equal(mapper.r2star_map().mean(), 1 / self.t2star)

        # Auto Threaded
        mapper = T2Star(signal_array, self.t, self.affine, method='2p_exp',
                        multithread='auto')
        assert mapper.shape == signal_array.shape[:-1]
        npt.assert_almost_equal(mapper.t2star_map.mean(), self.t2star)
        npt.assert_almost_equal(mapper.t2star_err.mean(), 7.395706644238e-11)
        npt.assert_almost_equal(mapper.m0_map.mean(), self.m0)
        npt.assert_almost_equal(mapper.r2star_map().mean(), 1 / self.t2star)

        # Fail to fit
        mapper = T2Star(signal_array[..., ::-1], self.t, self.affine,
                        method='2p_exp', multithread=True)
        assert mapper.shape == signal_array.shape[:-1]
        # Voxels that fail to fit are set to zero
        npt.assert_almost_equal(mapper.t2star_map.mean(), 0.0)
        npt.assert_almost_equal(mapper.t2star_err.mean(), 0.0)

    def test_mask(self):
        signal_array = np.tile(self.correct_signal, (10, 10, 3, 1))

        # Bool mask
        mask = np.ones(signal_array.shape[:-1], dtype=bool)
        mask[:5, :, :] = False
        mapper = T2Star(signal_array, self.t, self.affine, mask=mask)
        assert mapper.shape == signal_array.shape[:-1]
        npt.assert_almost_equal(mapper.t2star_map[5:, :, :].mean(),
                                self.t2star)
        npt.assert_almost_equal(mapper.t2star_map[:5, :, :].mean(), 0.0)

        # Int mask
        mask = np.ones(signal_array.shape[:-1])
        mask[:5, :, :] = 0
        mapper = T2Star(signal_array, self.t, self.affine, mask=mask)
        assert mapper.shape == signal_array.shape[:-1]
        npt.assert_almost_equal(mapper.t2star_map[5:, :, :].mean(),
                                self.t2star)
        npt.assert_almost_equal(mapper.t2star_map[:5, :, :].mean(), 0.0)

    def test_to_nifti(self):
        # Create a T1 map instance and test different export to NIFTI scenarios
        signal_array = np.tile(self.correct_signal, (10, 10, 3, 1))
        mapper = T2Star(signal_array, self.t, self.affine, method='2p_exp')

        os.makedirs('test_output', exist_ok=True)

        # Check all is saved.
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='t2startest', maps='all')
        output_files = os.listdir('test_output')
        assert len(output_files) == 6
        assert 't2startest_m0_err.nii.gz' in output_files
        assert 't2startest_m0_map.nii.gz' in output_files
        assert 't2startest_mask.nii.gz' in output_files
        assert 't2startest_r2star_map.nii.gz' in output_files
        assert 't2startest_t2star_err.nii.gz' in output_files
        assert 't2startest_t2star_map.nii.gz' in output_files

        for f in os.listdir('test_output'):
            os.remove(os.path.join('test_output', f))

        # Check that no files are saved.
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='t2startest', maps=[])
        output_files = os.listdir('test_output')
        assert len(output_files) == 0

        # Check that only t2star and r2star are saved.
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='t2startest', maps=['mask', 't2star',
                                                           'r2star'])
        output_files = os.listdir('test_output')
        assert len(output_files) == 3
        assert 't2startest_mask.nii.gz' in output_files
        assert 't2startest_r2star_map.nii.gz' in output_files
        assert 't2startest_t2star_map.nii.gz' in output_files

        for f in os.listdir('test_output'):
            os.remove(os.path.join('test_output', f))

        # Check that it fails when no maps are given
        with pytest.raises(ValueError):
            mapper = T2Star(signal_array, self.t, self.affine)
            mapper.to_nifti(output_directory='test_output',
                            base_file_name='t2startest', maps='')

        mapper = T2Star(signal_array, self.t, self.affine, method='loglin')

        # Check that error maps arent saved if method is `loglin`
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='t2startest', maps='all')
        output_files = os.listdir('test_output')
        assert len(output_files) == 4
        assert 't2startest_m0_map.nii.gz' in output_files
        assert 't2startest_mask.nii.gz' in output_files
        assert 't2startest_r2star_map.nii.gz' in output_files
        assert 't2startest_t2star_map.nii.gz' in output_files

        for f in os.listdir('test_output'):
            os.remove(os.path.join('test_output', f))

        # Check warning is produced if user explicitly asks for
        # t2star_err
        with pytest.warns(UserWarning):
            mapper.to_nifti(output_directory='test_output',
                            base_file_name='t2startest', maps=['t2star',
                                                               't2star_err'])
            output_files = os.listdir('test_output')
            print(output_files)
            assert len(output_files) == 2
            assert 't2startest_t2star_err.nii.gz' in output_files
            assert 't2startest_t2star_map.nii.gz' in output_files

        for f in os.listdir('test_output'):
            os.remove(os.path.join('test_output', f))

        # Check warning is produced if user explicitly asks for
        # t2star_err
        with pytest.warns(UserWarning):
            mapper.to_nifti(output_directory='test_output',
                            base_file_name='t2startest', maps=['m0', 'm0_err'])
            output_files = os.listdir('test_output')
            assert len(output_files) == 2
            assert 't2startest_m0_err.nii.gz' in output_files
            assert 't2startest_m0_map.nii.gz' in output_files

        # Delete 'test_output' folder
        shutil.rmtree('test_output')

    def test_missmatched_raw_data_and_echo_lengths(self):

        with pytest.raises(AssertionError):
            mapper = T2Star(pixel_array=np.zeros((5, 5, 4)),
                            echo_list=np.linspace(0, 2000, 5),
                            affine=self.affine)

        with pytest.raises(AssertionError):
            mapper = T2Star(pixel_array=np.zeros((5, 5, 5)),
                            echo_list=np.linspace(0, 2000, 4),
                            affine=self.affine)

    def test_methods(self):

        # Not a method string
        with pytest.raises(AssertionError):
            mapper = T2Star(pixel_array=np.zeros((5, 5, 5)),
                            echo_list=np.linspace(0, 2000, 5),
                            affine=self.affine, method='magic')

        # Int method
        with pytest.raises(AssertionError):
            mapper = T2Star(pixel_array=np.zeros((5, 5, 5)),
                            echo_list=np.linspace(0, 2000, 5),
                            affine=self.affine, method=0)

    def test_multithread_options(self):

        # Not valid string
        with pytest.raises(AssertionError):
            mapper = T2Star(pixel_array=np.zeros((5, 5, 5)),
                            echo_list=np.linspace(0, 2000, 5),
                            affine=self.affine, multithread='cloud')

    def test_loglin_warning(self):

        # Generate test data with T2* < 20 ms
        signal = two_param_eq(self.t, 10, self.m0)
        signal_array = np.tile(signal, (10, 10, 3, 1))
        with pytest.warns(UserWarning):
            mapper = T2Star(signal_array, self.t, affine=self.affine,
                            method='loglin')

    def test_real_data(self):

        # Get test data
        image, affine, te = fetch.t2star_philips()
        te *= 1000

        # Crop to reduce runtime
        image = image[30:60, 50:90, 2, :]

        # Gold standard statistics for each method
        gold_standard_loglin = [32.2660346964308, 18.499243841743308,
                                0.0, 239.07407841896983]
        gold_standard_2p_exp = [30.724443852557155, 22.156366883080896,
                                0.0, 529.8640757093401]

        # loglin method
        mapper = T2Star(image, te, self.affine, method='loglin')
        t2star_stats = arraystats.ArrayStats(mapper.t2star_map).calculate()
        npt.assert_allclose([t2star_stats["mean"], t2star_stats["std"],
                                    t2star_stats["min"], t2star_stats["max"]],
                                    gold_standard_loglin, rtol=1e-6, atol=1e-4)

        # 2p_exp method
        mapper = T2Star(image, te, self.affine, method='2p_exp')
        t2star_stats = arraystats.ArrayStats(mapper.t2star_map).calculate()
        npt.assert_allclose([t2star_stats["mean"], t2star_stats["std"],
                                    t2star_stats["min"], t2star_stats["max"]],
                                    gold_standard_2p_exp, rtol=1e-6, atol=1e-4)


# Delete the NIFTI test folder recursively if any of the unit tests failed
if os.path.exists('test_output'):
    shutil.rmtree('test_output')



import os
import shutil
import numpy as np
import numpy.testing as npt
import pytest
from ukat.data import fetch
from ukat.mapping.t2 import T2, two_param_eq, three_param_eq
from ukat.utils import arraystats


class TestT2:
    t2 = 120
    m0 = 3000
    t = np.linspace(12, 120, 10)
    b = 250

    # The ideal signal produced by the equation M0 * exp(-t / T2) where
    # M0 = 3000 and T2 = 120 ms at 10 echo times between 12 and 120 ms
    correct_signal = np.array([2714.51225411, 2456.19225923, 2222.45466205,
                               2010.96013811, 1819.59197914, 1646.43490828,
                               1489.75591137, 1347.98689235, 1219.70897922,
                               1103.63832351])
    affine = np.eye(4)

    def test_two_param_eq(self):
        signal = two_param_eq(self.t, self.t2, self.m0)
        npt.assert_allclose(signal, self.correct_signal, rtol=1e-6, atol=1e-8)

    def test_three_param_eq(self):
        signal = three_param_eq(self.t, self.t2, self.m0, self.b)
        npt.assert_allclose(signal, self.correct_signal + self.b,
                            rtol=1e-6, atol=1e-8)

    def test_2p_exp_fit(self):
        # Make the signal into a 4D array
        signal_array = np.tile(self.correct_signal, (10, 10, 3, 1))

        # Multithread
        mapper = T2(signal_array, self.t, self.affine, multithread=True)
        assert mapper.shape == signal_array.shape[:-1]
        npt.assert_almost_equal(mapper.t2_map.mean(), self.t2)
        npt.assert_almost_equal(mapper.m0_map.mean(), self.m0)
        npt.assert_almost_equal(mapper.r2_map().mean(), 1 / self.t2)

        # Single Threaded
        mapper = T2(signal_array, self.t, self.affine, multithread=False)
        assert mapper.shape == signal_array.shape[:-1]
        npt.assert_almost_equal(mapper.t2_map.mean(), self.t2)
        npt.assert_almost_equal(mapper.m0_map.mean(), self.m0)

        # Auto Threaded
        mapper = T2(signal_array, self.t, self.affine, multithread='auto')
        assert mapper.shape == signal_array.shape[:-1]
        npt.assert_almost_equal(mapper.t2_map.mean(), self.t2)
        npt.assert_almost_equal(mapper.m0_map.mean(), self.m0)

        # Fail to fit
        mapper = T2(signal_array[..., ::-1], self.t, self.affine,
                    multithread=True)
        assert mapper.shape == signal_array.shape[:-1]
        # Voxels that fail to fit are set to zero
        assert mapper.t2_map.mean() == 0.0

    def test_3p_exp_fit(self):
        # Make the signal into a 4D array
        signal_array = np.tile(self.correct_signal, (10, 10, 3, 1)) + self.b

        # Multithread
        mapper = T2(signal_array, self.t, self.affine, multithread=True,
                    method='3p_exp')
        assert mapper.shape == signal_array.shape[:-1]
        npt.assert_almost_equal(mapper.t2_map.mean(), self.t2)
        npt.assert_almost_equal(mapper.m0_map.mean(), self.m0)
        npt.assert_almost_equal(mapper.b_map.mean(), self.b)

        # Single Threaded
        mapper = T2(signal_array, self.t, self.affine, multithread=False,
                    method='3p_exp')
        assert mapper.shape == signal_array.shape[:-1]
        npt.assert_almost_equal(mapper.t2_map.mean(), self.t2)
        npt.assert_almost_equal(mapper.m0_map.mean(), self.m0)
        npt.assert_almost_equal(mapper.b_map.mean(), self.b)

    def test_threshold_fit(self):
        # Make the signal into a 4D array
        signal_array = np.tile(self.correct_signal, (10, 10, 3, 1))

        # Multithread
        mapper = T2(signal_array, self.t, self.affine, multithread=True,
                    noise_threshold=1300)
        assert mapper.shape == signal_array.shape[:-1]
        npt.assert_almost_equal(mapper.t2_map.mean(), self.t2)
        npt.assert_almost_equal(mapper.m0_map.mean(), self.m0)

        # Single Threaded
        mapper = T2(signal_array, self.t, self.affine, multithread=False,
                    noise_threshold=1300)
        assert mapper.shape == signal_array.shape[:-1]
        npt.assert_almost_equal(mapper.t2_map.mean(), self.t2)
        npt.assert_almost_equal(mapper.m0_map.mean(), self.m0)

    def test_mask(self):
        signal_array = np.tile(self.correct_signal, (10, 10, 3, 1))

        # Bool mask
        mask = np.ones(signal_array.shape[:-1], dtype=bool)
        mask[:5, :, :] = False
        mapper = T2(signal_array, self.t, self.affine, mask=mask)
        assert mapper.shape == signal_array.shape[:-1]
        npt.assert_almost_equal(mapper.t2_map[5:, :, :].mean(), self.t2)
        npt.assert_almost_equal(mapper.t2_map[:5, :, :].mean(), 0)

        # Int mask
        mask = np.ones(signal_array.shape[:-1])
        mask[:5, :, :] = 0
        mapper = T2(signal_array, self.t, self.affine, mask=mask)
        assert mapper.shape == signal_array.shape[:-1]
        npt.assert_almost_equal(mapper.t2_map[5:, :, :].mean(), self.t2)
        npt.assert_almost_equal(mapper.t2_map[:5, :, :].mean(), 0)

    def test_mismatched_raw_data_and_echo_lengths(self):
        with pytest.raises(AssertionError):
            mapper = T2(pixel_array=np.zeros((5, 5, 4)),
                        echo_list=np.linspace(0, 2000, 5),
                        affine=self.affine)

        with pytest.raises(AssertionError):
            mapper = T2(pixel_array=np.zeros((5, 5, 5)),
                        echo_list=np.linspace(0, 2000, 4),
                        affine=self.affine)

    def test_invalid_method(self):
        with pytest.raises(ValueError):
            mapper = T2(pixel_array=np.zeros((5, 5, 4)),
                        echo_list=np.linspace(0, 2000, 4),
                        affine=self.affine, method='4p_exp')

    def test_real_data(self):
        # Get test data
        image, affine, te = fetch.t2_philips(1)
        te *= 1000
        # Crop to reduce runtime
        image = image[60:90, 30:70, 2, :]

        # Gold standard statistics
        gold_standard_2p_exp = [105.63945, 39.616205,
                                0.0, 568.160604]
        gold_standard_3p_exp = [9.881218e+01, 4.294529e+01,
                                3.489657e-02, 5.681606e+02]
        gold_standard_thresh = [106.354968,  39.894933,
                                0.0, 568.160591]

        # 2p_exp method
        mapper = T2(image, te, self.affine)
        t2_stats = arraystats.ArrayStats(mapper.t2_map).calculate()
        npt.assert_allclose([t2_stats["mean"], t2_stats["std"],
                             t2_stats["min"], t2_stats["max"]],
                            gold_standard_2p_exp, rtol=1e-6, atol=1e-4)

        # 3p_exp method
        mapper = T2(image, te, self.affine, method='3p_exp')
        t2_stats = arraystats.ArrayStats(mapper.t2_map).calculate()
        npt.assert_allclose([t2_stats["mean"], t2_stats["std"],
                             t2_stats["min"], t2_stats["max"]],
                            gold_standard_3p_exp, rtol=1e-6, atol=1e-4)

        # threshold method
        mapper = T2(image, te, self.affine, noise_threshold=100)
        t2_stats = arraystats.ArrayStats(mapper.t2_map).calculate()
        npt.assert_allclose([t2_stats["mean"], t2_stats["std"],
                             t2_stats["min"], t2_stats["max"]],
                            gold_standard_thresh, rtol=1e-6, atol=1e-4)

    def test_to_nifti(self):
        # Create a T2 map instance and test different export to NIFTI scenarios
        signal_array = np.tile(self.correct_signal, (10, 10, 3, 1))
        mapper = T2(signal_array, self.t, self.affine)

        os.makedirs('test_output', exist_ok=True)

        # Check all is saved.
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='t2test', maps='all')
        output_files = os.listdir('test_output')
        assert len(output_files) == 6
        assert 't2test_m0_err.nii.gz' in output_files
        assert 't2test_m0_map.nii.gz' in output_files
        assert 't2test_mask.nii.gz' in output_files
        assert 't2test_r2_map.nii.gz' in output_files
        assert 't2test_t2_err.nii.gz' in output_files
        assert 't2test_t2_map.nii.gz' in output_files

        for f in os.listdir('test_output'):
            os.remove(os.path.join('test_output', f))

        # Check that no files are saved.
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='t2test', maps=[])
        output_files = os.listdir('test_output')
        assert len(output_files) == 0

        # Check that only t2 and r2 are saved.
        mapper.to_nifti(output_directory='test_output',
                        base_file_name='t2test', maps=['mask', 't2', 'r2'])
        output_files = os.listdir('test_output')
        assert len(output_files) == 3
        assert 't2test_mask.nii.gz' in output_files
        assert 't2test_t2_map.nii.gz' in output_files
        assert 't2test_r2_map.nii.gz' in output_files

        for f in os.listdir('test_output'):
            os.remove(os.path.join('test_output', f))

        # Check that it fails when no maps are given
        with pytest.raises(ValueError):
            mapper = T2(signal_array, self.t, self.affine)
            mapper.to_nifti(output_directory='test_output',
                            base_file_name='t2test', maps='')

        # Delete 'test_output' folder
        shutil.rmtree('test_output')


# Delete the NIFTI test folder recursively if any of the unit tests failed
if os.path.exists('test_output'):
    shutil.rmtree('test_output')



from setuptools import setup, find_packages

# Get requirements from text file
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

# Use README.md as the long description
with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="ukat",
    version="0.6.2",
    description="UKRIN Kidney Analysis Toolbox",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/UKRIN-MAPS/ukat",
    license="GPL-3.0",

    python_requires='>=3.8, <4',
    packages=find_packages(),
    install_requires=requirements,
    include_package_data=True,

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Environment :: Console',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    ],
)
